/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * CPIface text track/pattern display
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -fd981119   Felix Domke <tmbinc@gmx.net>
 *    -added the really important 'NO_CPIFACE_IMPORT'
 *  -ss04????   Stian Skjelstad <stian@nixia.no>
 *    -many changes to make it resize to any-sizes more pretty
 *  -ss040915   Stian Skjelstad <stian@nixia.no>
 *    -Only print the right row-numbers if it fits without it makes us drop more
 *     rows
 *  -ss040918   Stian Skjelstad <stian@nixia.no>
 *    -Fixed a bug when scrolling channels, when they are zoomed wider than the screen
 */

#define NO_CPIFACE_IMPORT
#include "config.h"
#include <string.h>
#include <stdlib.h>
#include "types.h"
#include "stuff/poutput.h"
#include "boot/psetting.h"
#include "cpiface.h"

#include <curses.h>

#define COLPTNOTE 0x0A
#define COLNOTE 0x0F
#define COLPITCH 0x02
#define COLSPEED 0x02
#define COLPAN 0x05
#define COLVOL 0x09
#define COLACT 0x04
#define COLINS 0x07
#define COLLNUM 0x07
#define COLHLNUM 0x04
#define COLBACK 0x08
#define COLMUTE 0x08
#define COLTITLE1 0x01
#define COLTITLE1H 0x09
#define COLTITLE2 0x07
#define COLTITLE2M 0x08
#define COLTITLE2H 0x0F
#define COLMARK 0x0F
#define COLHLAND 0xFF
#define COLHLOR 0x88

static int plTrackActive;

static int (*getcurpos)();
static int (*getpatlen)(int n);
static const char *(*getpatname)(int n);
static void (*seektrack)(int n, int c);
static int (*startrow)();
static int (*getnote)(unsigned short *bp, int small);
static int (*getins)(unsigned short *bp);
static int (*getvol)(unsigned short *bp);
static int (*getpan)(unsigned short *bp);
static void (*getfx)(unsigned short *bp, int n);
static void (*getgcmd)(unsigned short *bp, int n);

static int plPatternNum;
static short plPrepdPat;
static short plPatType;
static short plPatFirstLine;
static short plPatHeight;
static short plPatWidth;
static short plPatManualPat;
static short plPatManualRow;
static unsigned short (*plPatBuf)[CONSOLE_MAX_X];
static unsigned short pathighlight[CONSOLE_MAX_X];
static char pattitle1[CONSOLE_MAX_X+1];
static unsigned short pattitle2[CONSOLE_MAX_X];
static short patwidth, patpad;

enum
{
	cpiTrkFXIns=1,cpiTrkFXNote=2,cpiTrkFXVol=4,cpiTrkFXNoPan=8
};

static void getfx2(unsigned short *bp, int n, int o)
{
	int p=0;
	if (o&cpiTrkFXIns)
		if (getins(bp+1))
		{
			writestring(bp, 0, COLINS, "i", 1);
			p++;
			bp+=3;
		}
	if (p==n)
		return;
	if (o&cpiTrkFXNote)
		if (getnote(bp, 0))
		{
			p++;
			bp+=3;
		}
	if (p==n)
		return;
	if (o&cpiTrkFXVol)
		if (getvol(bp+1))
		{
			writestring(bp, 0, COLVOL, "v", 1);
			p++;
			bp+=3;
		}
	if (p==n)
		return;
	if (!(o&cpiTrkFXNoPan))
		if (getpan(bp+1))
		{
			writestring(bp, 0, COLPAN, "p", 1);
			p++;
			bp+=3;
		}
	if (p==n)
		return;
	getfx(bp, n-p);
}

static void getscrollpos(int tr, int *firstchan, int *chnn)
{
	if (plNLChan>tr)
	{
		if (plSelCh<(tr>>1))
			*firstchan=0;
		else
			if (plSelCh>=(plNLChan-(tr>>1)))
				*firstchan=plNLChan-tr;
			else
				*firstchan=plSelCh-((tr>>1));
		*chnn=tr;
	} else {
		*firstchan=0;
		*chnn=plNLChan;
	}
}


static void setattrgrey(unsigned short *buf, unsigned short len)
{
	short i;
	for (i=0; i<len; i++)
	{
		((char*)buf)[1]=COLMUTE;
		buf++;
	}
}

static void preparetrack1(unsigned short *bp)
{
	getnote(bp, 2);
}

static void preparetrack2(unsigned short *bp)
{
	getnote(bp, 1);
}

static void preparetrack3(unsigned short *bp)
{
	getnote(bp, 0);
}

static void preparetrack3f(unsigned short *bp)
{
	if (!getnote(bp, 0))
		getfx2(bp, 1, cpiTrkFXVol);
}


static void preparetrack6nf(unsigned short *bp)
{
	getnote(bp, 0);
	getfx2(bp+3, 1, cpiTrkFXVol);
}

static void preparetrack17invff(unsigned short *bp)
{
	getins(bp);
	getnote(bp+3, 0);
	getvol(bp+7);
	getfx2(bp+10, 2, 0);
}

static void preparetrack14invff(unsigned short *bp)
{
	getins(bp);
	getnote(bp+2, 0);
	getvol(bp+5);
	getfx2(bp+7, 2, 0);
}

static void preparetrack14nvff(unsigned short *bp)
{
	getnote(bp, 0);
	getvol(bp+4);
	getfx2(bp+7, 2, 0);
}

static void preparetrack26invpffff(unsigned short *bp)
{
	getins(bp);
	getnote(bp+3, 0);
	getvol(bp+7);
	getpan(bp+10);
	getfx2(bp+13, 4, cpiTrkFXNoPan);
}

static void preparetrack16fffff(unsigned short *bp)
{
	getfx2(bp, 5, cpiTrkFXIns|cpiTrkFXNote|cpiTrkFXVol);
}

static void preparetrack8nvf(unsigned short *bp)
{
	getnote(bp, 0);
	getvol(bp+3);
	getfx2(bp+5, 1, 0);
}

static void preparetrack8inf(unsigned short *bp)
{
	getins(bp);
	getnote(bp+2, 0);
	getfx2(bp+5, 1, cpiTrkFXVol);
}

struct patviewtype
{
/*	uint8_t maxch;   These are now calculated real-time */
	uint8_t gcmd;
	uint8_t width;
	const char *title;
	const char *paused;
	const char *normal;
	const char *selected;
	void (*putcmd)(unsigned short *bp);
};

static void preparepatgen(int pat, const struct patviewtype *pt)
{
	int i;
	int plen;
	int firstchan,chnn;
	const char *pname;
	short p0, n0;
	unsigned short patmask[CONSOLE_MAX_X];
	int firstrow=20;
	int firstprow=0;
	int firstpat;

	int maxch=(plPatWidth-pt->gcmd*4-3)/pt->width;

	if (maxch>plNLChan)
		maxch=plNLChan;

	patpad=(plPatWidth-maxch*pt->width-pt->gcmd*4-3)>=4;

	plPrepdPat=pat;
	plen=getpatlen(pat);

	getscrollpos(maxch, &firstchan, &chnn);

	strcpy(pattitle1, "   pattern view:  order ");
	convnum(pat, pattitle1+strlen(pattitle1), 16, 3, 0);
	strcat(pattitle1, ", ");
	convnum(maxch, pattitle1+strlen(pattitle1), 10, 2, 1);
	strcat(pattitle1, " channels,  ");
	strcat(pattitle1, pt->title);
	pname=getpatname(pat);
	if (pname&&*pname)
	{
		strcat(pattitle1, ": ");
		strcat(pattitle1, pname);
	}

	p0=4+4*pt->gcmd;
	patwidth=p0+pt->width*chnn+4;
	writestring(pattitle2, 0, COLTITLE2, "row", CONSOLE_MAX_X);
	if (patpad)
		writestring(pattitle2, patwidth-3, COLTITLE2, "row", 3);
	switch (pt->gcmd)
	{
		case 0:
			break;
		case 1:
			writestring(pattitle2, 4, plPause?COLTITLE2M:COLTITLE2, "gbl", 3);
			break;
		case 2:
			writestring(pattitle2, 5, plPause?COLTITLE2M:COLTITLE2, "global", 6);
			break;
		case 3:
			writestring(pattitle2, 5, plPause?COLTITLE2M:COLTITLE2, "global cmd", 10);
			break;
		default:
			writestring(pattitle2, pt->gcmd*2-4, plPause?COLTITLE2M:COLTITLE2, "global commands", 15);
			break;
	}


	writestring(patmask, 0, COLLNUM, "00 ", CONSOLE_MAX_X);
	if (patpad)
		writestring(patmask, patwidth-3, COLLNUM, "00", 2);
	writestring(patmask, 3, COLBACK, "", 1);
	writestring(patmask, 3+pt->gcmd*4, COLBACK, "", 1);
	if (!plPause)
		for (i=0; i<pt->gcmd; i++)
			writestring(patmask, 4+4*i, COLBACK, "", 3);

	n0=p0+(pt->width+1)/2-1;
	if (pt->width==4)
		n0--;
	for (i=0; i<chnn; i++)
	{
		char chpaus=plMuteCh[i+firstchan];
		char sel=((i+firstchan)==plSelCh);
		writenum(pattitle2, n0+pt->width*i, sel?COLTITLE2H:chpaus?COLTITLE2M:COLLNUM, i+firstchan+1, 10, (pt->width==1)?1:2, pt->width>2);
		writestring(patmask, p0+pt->width*i, COLBACK, chpaus?pt->paused:sel?pt->selected:pt->normal, pt->width);
	}

	firstpat=pat;

	while (firstpat)
	{
		firstpat--;
		firstrow-=getpatlen(firstpat);
		if (firstrow<=0)
		{
			firstprow-=firstrow;
			firstrow=0;
			break;
		}
	}
	for (i=0; i<firstrow; i++)
		writestring(plPatBuf[i], 0, COLBACK, "", CONSOLE_MAX_X);

	while (firstpat<plPatternNum)
	{
		int curlen;
		int lastprow;

		if (!getpatlen(firstpat))
		{
			firstpat++;
			continue;
		}
		curlen=getpatlen(firstpat);
		lastprow=curlen;
		if ((firstrow+curlen-firstprow)>(plen+60))
			lastprow=plen+60-firstrow+firstprow;

		for (i=firstprow; i<lastprow; i++)
		{
			writestringattr(plPatBuf[i+firstrow-firstprow], 0, patmask, CONSOLE_MAX_X);
			writenum(plPatBuf[i+firstrow-firstprow], 0, i?COLLNUM:COLHLNUM, i, 16, 2, 0);
			if (patwidth<=132)
			writenum(plPatBuf[i+firstrow-firstprow], patwidth-3, i?COLLNUM:COLHLNUM, i, 16, 2, 0);
		}

		if (pt->gcmd)
		{
			seektrack(firstpat, -1);
			while (1)
			{
				int currow=startrow();
				if (currow==-1)
					break;
				if ((currow>=firstprow)&&(currow<lastprow))
				{
					unsigned short *bp=plPatBuf[currow+firstrow-firstprow]+4;
					getgcmd(bp, pt->gcmd);
					if (plPause)
						setattrgrey(bp, pt->gcmd*4);
				}
			}
		}

		for (i=0; i<chnn; i++)
		{
			int chpaus;
			seektrack(firstpat, i+firstchan);
			chpaus=plMuteCh[i+firstchan];
			while (1)
			{
				int currow=startrow();
				if (currow==-1)
					break;
				if ((currow>=firstprow)&&(currow<lastprow))
				{
					unsigned short *bp=plPatBuf[currow+firstrow-firstprow]+p0+i*pt->width;
					pt->putcmd(bp);
					if (chpaus)
						setattrgrey(bp, pt->width);
				}
			}
		}

		firstrow+=lastprow-firstprow;
		firstprow=0;
		firstpat++;

		if (firstrow==(plen+60))
			break;
	}

	for (i=firstrow; i<(plen+60); i++)
		writestring(plPatBuf[i], 0, COLBACK, "", CONSOLE_MAX_X);
}

static struct patviewtype pat6480   = {/*64,*/  3,  1, "(notes only)",                          " ",
                                                                                                "",
                                                                                                "", preparetrack1};
static struct patviewtype pat64132  = {/*64,*/ 15,  1, "(notes only)",                          " ",
                                                                                                "",
                                                                                                "", preparetrack1};
static struct patviewtype pat64132m = {/*64,*/  0,  2, "(notes only)",                          "  ",
                                                                                                "",
                                                                                                "", preparetrack2};
static struct patviewtype pat4880   = {/*48,*/  6,  1, "(notes only)",                          " ",
                                                                                                "",
                                                                                                "", preparetrack1};
static struct patviewtype pat48132  = {/*48,*/  7,  2, "(notes only)",                          "  ",
                                                                                                "",
                                                                                                "", preparetrack2};
static struct patviewtype pat3280   = {/*32,*/  3,  2, "(notes only)",                          "  ",
                                                                                                "",
                                                                                                "", preparetrack2};
static struct patviewtype pat32132  = {/*32,*/  7,  3, "(notes only)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3};
static struct patviewtype pat32132f = {/*32,*/  7,  3, "(notes & fx)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3f};
static struct patviewtype pat2480   = {/*24,*/  1,  3, "(notes only)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3};
static struct patviewtype pat2480f  = {/*24,*/  1,  3, "(notes & fx)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3f};
static struct patviewtype pat24132  = {/*24,*/  7,  4, "(notes only)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3};
static struct patviewtype pat24132f = {/*24,*/  7,  4, "(notes & fx)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3f};
static struct patviewtype pat1680   = {/*16,*/  3,  4, "(notes only)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3};
static struct patviewtype pat1680f  = {/*16,*/  3,  4, "(notes & fx)",                          "   ",
                                                                                                "",
                                                                                                "", preparetrack3f};
static struct patviewtype pat16132  = {/*16,*/  3,  7, "(note, fx)",                            "      ",
                                                                                                "   ",
                                                                                                "", preparetrack6nf};
static struct patviewtype pat880    = {/* 8,*/  1,  9, "(ins, note, fx)",                       "        ",
                                                                                                "     ",
                                                                                                "", preparetrack8inf};
static struct patviewtype pat880f   = {/* 8,*/  1,  9, "(note, vol, fx)",                       "        ",
                                                                                                "  ",
                                                                                                "", preparetrack8nvf};
static struct patviewtype pat8132   = {/* 8,*/  3, 14, "(note, vol, fx, fx)",                   "             ",
                                                                                                "  ",
                                                                                                "  ", preparetrack14nvff};
static struct patviewtype pat8132f  = {/* 8,*/  3, 14, "(ins, note, vol, fx, fx)",              "             ",
                                                                                                "    ",
                                                                                                "", preparetrack14invff};
static struct patviewtype pat480    = {/* 4,*/  2, 17, "(ins, note, vol, fx, fx)",              "                ",
                                                                                                "   ",
                                                                                                "   ", preparetrack17invff};
static struct patviewtype pat480f   = {/* 4,*/  3, 16, "(fx, fx, fx, fx, fx)",                  "               ",
                                                                                                "",
                                                                                                "", preparetrack16fffff};
static struct patviewtype pat4132   = {/* 4,*/  5, 26, "(ins, note, vol, pan, fx, fx, fx, fx)", "                         ",
	                                                                                        "    ",
                                                                                                "    ", preparetrack26invpffff};

static int max(int a, int b)
{
	if (a>b)
		return a;
	else
		return b;
}

static void gmdDrawPattern(char sel)
{
	int pos=getcurpos();
	short pat=pos>>8;
	int currow=pos&0xFF;
	int curpat=pat;
	short crow=currow;
	int i,j,row;
	int plen;

	if (plPatManualPat!=-1)
	{
		pat=plPatManualPat;
		crow=plPatManualRow;
	}
	while (!getpatlen(pat))
	{
		pat++;
		crow=0;
		if (pat>=plPatternNum)
			pat=0;
	}

	if ((plPrepdPat!=pat)||plChanChanged)
	{
		if (plPatWidth<128)
			switch (plPatType)
			{
				case 0: preparepatgen(pat, &pat6480); break;
				case 1: preparepatgen(pat, &pat6480); break;
				case 2: preparepatgen(pat, &pat4880); break;
				case 3: preparepatgen(pat, &pat4880); break;
				case 4: preparepatgen(pat, &pat3280); break;
				case 5: preparepatgen(pat, &pat3280); break;
				case 6: preparepatgen(pat, &pat2480); break;
				case 7: preparepatgen(pat, &pat2480f); break;
				case 8: preparepatgen(pat, &pat1680); break;
				case 9: preparepatgen(pat, &pat1680f); break;
				case 10: preparepatgen(pat, &pat880); break;
				case 11: preparepatgen(pat, &pat880f); break;
				case 12: preparepatgen(pat, &pat480f); break;
				case 13: preparepatgen(pat, &pat480); break;
			} else switch (plPatType)
			{
				case 0: preparepatgen(pat, &pat64132m); break;
				case 1: preparepatgen(pat, &pat64132); break;
				case 2: preparepatgen(pat, &pat48132); break;
				case 3: preparepatgen(pat, &pat48132); break;
				case 4: preparepatgen(pat, &pat32132); break;
				case 5: preparepatgen(pat, &pat32132f); break;
				case 6: preparepatgen(pat, &pat24132); break;
				case 7: preparepatgen(pat, &pat24132f); break;
				case 8: preparepatgen(pat, &pat16132); break;
				case 9: preparepatgen(pat, &pat16132); break;
				case 10: preparepatgen(pat, &pat8132f); break;
				case 11: preparepatgen(pat, &pat8132); break;
				case 12: preparepatgen(pat, &pat4132); break;
				case 13: preparepatgen(pat, &pat4132); break;
			}
	}

	/* show the two header lines*/
	displaystr(plPatFirstLine-2, 0, sel?COLTITLE1H:COLTITLE1, pattitle1, plPatWidth);
	displaystrattr(plPatFirstLine-1, 0, pattitle2, plPatWidth);
	/* done */

	plen=getpatlen(pat);
	for (i=0, row=max(crow-plPatHeight/3, -20); i<plPatHeight; i++, row++)
	{
		if ((row!=crow)&&((plPatManualPat==-1)||(row!=currow)||(pat!=curpat)))
		{
			displaystrattr(plPatFirstLine+i, 0, plPatBuf[20+row], plPatWidth);
			continue;
		}
		writestringattr(pathighlight, 0, plPatBuf[20+row], plPatWidth);
		if ((row==currow)&&(pat==curpat))
		{
			writestring(pathighlight, 2, COLMARK, "\x10", 1);
			if (patwidth>=132)
				writestring(pathighlight, patwidth-4, COLMARK, "\x11", 1);
		}
		if (row==crow)
			for (j=0; j<patwidth; j++)
				pathighlight[j]=(pathighlight[j]|(COLHLOR<<8))&((COLHLAND<<8)|0xFF);
		displaystrattr(plPatFirstLine+i, 0, pathighlight, plPatWidth);
	}
}

static int gmdTrkProcessKey(unsigned short key)
{
	switch (key)
	{
		case KEY_PPAGE:
			if (plPatManualPat==-1)
			{
				if (plPatType>1)
				{
					plPatType-=2;
					plPrepdPat=-1;
				}
			} else {
				plPatManualRow-=8;
				if (plPatManualRow<0)
				{
					plPatManualPat--;
					if (plPatManualPat<0)
						plPatManualPat=plPatternNum-1;
					while (!getpatlen(plPatManualPat))
						plPatManualPat--;
					plPatManualRow=getpatlen(plPatManualPat)-1;
				}
			}
			break;
		case KEY_NPAGE:
			if (plPatManualPat==-1)
			{
				if (plPatType<12)
				{
					plPatType+=2;
					plPrepdPat=-1;
				}
			} else {
				plPatManualRow+=8;
				if (plPatManualRow>=getpatlen(plPatManualPat))
				{
					plPatManualPat++;
					while ((plPatManualPat<plPatternNum)&&!getpatlen(plPatManualPat))
						plPatManualPat++;
					if (plPatManualPat>=plPatternNum)
						plPatManualPat=0;
					plPatManualRow=0;
				}
			}
			break;
/* ALREADY COMMENTED
  case 0x8400: //ctrl-pgup
    if (plInstType)
      plInstScroll--;
    break;
  case 0x7600: //ctrl-pgdn
    if (plInstType)
      plInstScroll++;
    break;
*/
		case KEY_HOME:
			plPatType=(plNLChan<=4)?13:(plNLChan<=8)?11:(plNLChan<=16)?9:(plNLChan<=24)?7:(plNLChan<=32)?5:(plNLChan<=48)?3:1;
/* TODO   winrecalc(); */
			break;
		case 9: /* tab */
			if (plPatManualPat==-1)
			{
				plPatType^=1;
				plPrepdPat=-1;
			} else {
				if (plPatType<13)
				{
					plPatType++;
					plPrepdPat=-1;
				}
			}
			break;
/* TODO-keys	case 0x0F00: // shift-tab
		case 0xA500:
			if (plPatManualPat==-1)
			{
				plPatType^=1;
				plPrepdPat=-1;
			} else {
				if (plPatType)
				{
					plPatType--;
					plPrepdPat=-1;
				}
			}
			break;*/
		case ' ':
			if (plPatManualPat==-1)
			{
				int p=getcurpos();
				plPatManualPat=p>>8;
				plPatManualRow=p&0xFF;
			} else
				plPatManualPat=-1;
			break;
		default:
			return 0;
	}
	return 1;
}

static void TrakSetWin(int _ignore, int wid, int ypos, int hgt)
{
	plPatFirstLine=ypos+2;
	plPatHeight=hgt-2;
	plPatWidth=wid;
}

static int TrakGetWin(struct cpitextmodequerystruct *q)
{
	if (!plTrackActive)
		return 0;

	q->hgtmin=3;
	q->hgtmax=100;
	q->xmode=1;
	q->size=2;
	q->top=0;
	q->killprio=64;
	q->viewprio=160;
	return 1;
}

static void TrakDraw(int focus)
{
	gmdDrawPattern(focus);
}

static int TrakIProcessKey(unsigned short key)
{
	switch (key)
	{
		case 't': case 'T':
			cpiTextSetMode("trak");
			break;
		case 'x': case 'X':
			plTrackActive=1;
			return 0;
/* TODO-keys	case 0x2d00: //alt-x
			plTrackActive=0;
			return 0;*/
		default:
			return 0;
	}
	return 1;
}

static int TrakAProcessKey(unsigned short key)
{
	switch (key)
	{
		case 't': case 'T':
			plTrackActive=!plTrackActive;
			cpiTextRecalc();
			break;
		default:
			return gmdTrkProcessKey(key);
	}
	return 1;
}

static int trkEvent(int ev)
{
	switch (ev)
	{
		case cpievInitAll:
			plTrackActive=cfGetProfileBool2(cfScreenSec, "screen", "pattern", 1, 1);
			return 0;
		case cpievInit:
			plPatBuf=calloc(sizeof(short), (256+60)*CONSOLE_MAX_X);
			if (!plPatBuf)
				return 0;
			break;
		case cpievDone:
			free(plPatBuf);
			break;
	}
	return 1;
}

struct cpitextmoderegstruct cpiTModeTrack = {"trak", TrakGetWin, TrakSetWin, TrakDraw, TrakIProcessKey, TrakAProcessKey, trkEvent};

void cpiTrkSetup(const struct cpitrakdisplaystruct *c, int npat)
{
	plPatternNum=npat;
	plPatManualPat=-1;
	plPrepdPat=-1;
	plPatType=(plNLChan<=4)?13:(plNLChan<=8)?11:(plNLChan<=16)?9:(plNLChan<=24)?7:(plNLChan<=32)?5:(plNLChan<=48)?3:1;
	getcurpos=c->getcurpos;
	getpatlen=c->getpatlen;
	getpatname=c->getpatname;
	seektrack=c->seektrack;
	startrow=c->startrow;
	getnote=c->getnote;
	getins=c->getins;
	getvol=c->getvol;
	getpan=c->getpan;
	getfx=c->getfx;
	getgcmd=c->getgcmd;
	cpiTextRegisterMode(&cpiTModeTrack);
}
