/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Wavetable devices system
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -kb980717   Tammo Hinrichs <opencp@gmx.net>
 *    -changed INI reading of driver symbols to _dllinfo lookup
 */

#define NO_MCPBASE_IMPORT

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "types.h"
#include "filesel/mdb.h"
#include "filesel/modlist.h"
#include "filesel/pfilesel.h"
#include "imsdev.h"
#include "player.h"
#include "boot/psetting.h"
#include "boot/plinkman.h"
#include "stuff/err.h"
#include "boot/pmain.h"
#include "devigen.h"
#include "mcp.h"
#include "stuff/compat.h"

int (*mcpProcessKey)(unsigned short);

struct devinfonode *plWaveTableDevices;
static struct devinfonode *curwavedev;
static struct devinfonode *defwavedev;

static struct devinfonode *getdevstr(struct devinfonode *n, const char *hnd)
{
	while (n)
	{
		if (!strcasecmp(n->handle, hnd))
			return n;
		n=n->next;
	}
	return 0;
}

static void setdevice(struct devinfonode **curdev, struct devinfonode *dev)
{
	if (*curdev==dev)
		return;
	if (*curdev)
	{
		if ((*curdev)->addprocs)
			if ((*curdev)->addprocs->Close)
				(*curdev)->addprocs->Close();
		mcpProcessKey=0;
		(*curdev)->dev.dev->Close();
		if (!(*curdev)->keep)
		{
			lnkFree((*curdev)->linkhand);
			(*curdev)->linkhand=-1;
		}
	}
	*curdev=0;
	if (!dev)
		return;

	if (dev->linkhand<0)
	{
		char lname[12];
		char *dsym;
		strncpy(lname,cfGetProfileString(dev->handle, "link", ""),11);
		dev->linkhand=lnkLink(lname);
		if (dev->linkhand<0)
		{
			fprintf(stderr, "device load error\n");
			return;
		}
		dev->dev.dev=(struct sounddevice *)_lnkGetSymbol(lnkReadInfoReg(dev->linkhand, "driver"));
		if (!dev->dev.dev)
		{
			fprintf(stderr, "device symbol error\n");
			lnkFree(dev->linkhand);
			dev->linkhand=-1;
			return;
		}
		dsym=lnkReadInfoReg(dev->linkhand, "addprocs");
	
		dev->addprocs=0;	
		if (dsym)
			dev->addprocs=0;
		else
			dev->addprocs=(struct devaddstruct *)_lnkGetSymbol(dsym);
	}

	fprintf(stderr, "%s selected...\n", dev->name);
	if (dev->dev.dev->Init(&dev->dev))
	{
		if (dev->addprocs)
			if (dev->addprocs->Init)
				dev->addprocs->Init(dev->handle);
		if (dev->addprocs)
			if (dev->addprocs->ProcessKey)
				mcpProcessKey=dev->addprocs->ProcessKey;
		*curdev=dev;
		return;
	}
	if (!(*curdev)->keep)
	{
		lnkFree((*curdev)->linkhand);
		(*curdev)->linkhand=-1;
	}
	fprintf(stderr, "device init error\n");
}


void mcpSetDevice(const char *name, int def)
{
	setdevice(&curwavedev, getdevstr(plWaveTableDevices, name));
	if (def)
		defwavedev=curwavedev;
}

void mcpResetDevice(void)
{
	setdevice(&curwavedev, defwavedev);
}

static int wavedevinit(void)
{
	const char *def;
	int playrate;

#ifdef INITCLOSE_DEBUG
	fprintf(stderr, "wavedevinit... trying to find all wavetables    [sound]->wavetabledevices\n");
#endif

	if (!strlen(cfGetProfileString2(cfSoundSec, "sound", "wavetabledevices", "")))
		return errOk;
	fprintf(stderr, "wavetabledevices:\n");
	if (!deviReadDevices(cfGetProfileString2(cfSoundSec, "sound", "wavetabledevices", ""), &plWaveTableDevices))
	{
		fprintf(stderr, "could not install wavetable devices!\n");
		return errGen;
	}

	curwavedev=0;
	defwavedev=0;

	def=cfGetProfileString("commandline_s", "w", cfGetProfileString2(cfSoundSec, "sound", "defwavetable", ""));

	if (strlen(def))
		mcpSetDevice(def, 1);
	else
		if (plWaveTableDevices)
			mcpSetDevice(plWaveTableDevices->handle, 1);

	fprintf(stderr, "\n");

	playrate=cfGetProfileInt("commandline_s", "r", cfGetProfileInt2(cfSoundSec, "sound", "mixrate", 44100, 10), 10);
	if (playrate<66)
	{
		if (playrate%11)
			playrate*=1000;
		else
			playrate=playrate*11025/11;
	}
	mcpMixOpt=0;
	if (!cfGetProfileBool("commandline_s", "8", !cfGetProfileBool2(cfSoundSec, "sound", "mix16bit", 1, 1), 1))
		mcpMixOpt|=PLR_16BIT;
	if (!cfGetProfileBool("commandline_s", "m", !cfGetProfileBool2(cfSoundSec, "sound", "mixstereo", 1, 1), 1))
		mcpMixOpt|=PLR_STEREO;
	mcpMixMaxRate=playrate;
	mcpMixProcRate=cfGetProfileInt2(cfSoundSec, "sound", "mixprocrate", 1536000, 10);
	mcpMixBufSize=cfGetProfileInt2(cfSoundSec, "sound", "mixbufsize", 100, 10)*65;
	mcpMixPoll=mcpMixBufSize;
	mcpMixMax=mcpMixBufSize;

	return errOk;
}

static void wavedevclose(void)
{
#ifdef INITCLOSE_DEBUG
	fprintf(stderr, "wavedevclose....\n");
#endif

	setdevice(&curwavedev, 0);
	while (plWaveTableDevices)
	{
		struct devinfonode *o=plWaveTableDevices;
		plWaveTableDevices=plWaveTableDevices->next;
		free(o);
	}
}

static int mcpReadDir(struct modlist *ml, const char *drive, const char *path, const char *mask, unsigned long opt)
{
	struct modlistentry m;
/*  if (opt&RD_PUTDSUBS)
  {
    fsConvFileName12(m.name, "@:", "");
    m.fileref=0xFFFF;
    m.dirref=dmGetDriveDir(0);
    if (!mdbAppendNew(ml, m))
      return 0;
  }*/
	if (strcmp(drive, "setup:"))
		return 1;

	if (opt&RD_PUTSUBS)
	{
		if (!strcmp(path, "/"))
		if ((ml->find(ml, "/DEVICES/"))<0)
		{
/*			m=calloc(1, sizeof(struct modlistentry));*/
			memset(&m, 0, sizeof(m));

			m.drive=drive;
			strcpy(m.name, "DEVICES");
			strcpy(m.shortname, "DEVICES");
			strcpy(m.fullname, "/DEVICES/");
			m.flags=MODLIST_FLAG_DIR;
			ml->append(ml, &m);
		}
			
	}
/*
  if (opt&RD_PUTSUBS)
  {
    dmGetPath(path, dirref);
    if (!strcmp(path, "@:\\"))
    {
      fsConvFileName12(m.name, "DEVICES", "");
      dmGetPath(path, dirref);
      strcat(path, "DEVICES");
      m.dirref=dmGetPathReference(path);
      if (m.dirref==0xFFFF)
        return 0;
      m.fileref=0xFFFE;
      if (!mdbAppendNew(ml, m))
        return 0;
    }
  }
*/
	if (!strcmp(path, "/DEVICES/"))

	{
		struct devinfonode *dev;
		for (dev=plWaveTableDevices; dev; dev=dev->next)
		{
			char hnd[9];
			strcpy(hnd, dev->handle);
/*			strupr(hnd);*/
			memset(&m, 0, sizeof(m));

			fsConvFileName12(m.name, hnd, ".DEV");
/*			if (fsMatchFileName12(m.name, mask))*/
			{
				m.fileref=mdbGetModuleReference(m.name, dev->dev.mem);
				if (m.fileref==0xffffffff)
					return 0;
				m.drive=drive;
/*				m.dirref=dirref;*/
				strncpy(m.shortname, m.name, 12);
				strcpy(m.fullname, "/DEVICES/");
				strcat(m.fullname, hnd);
				strcat(m.fullname, ".DEV");
				m.flags=MODLIST_FLAG_FILE|MODLIST_FLAG_VIRTUAL;
				if (mdbGetModuleType(m.fileref)!=mtDEVw)
				{
					struct moduleinfostruct mi;
					mdbGetModuleInfo(&mi, m.fileref);
					mi.flags1|=MDB_VIRTUAL;
					mi.channels=dev->dev.chan;
					strcpy(mi.modname, dev->name);
					mi.modtype=mtDEVw;
					mdbWriteModuleInfo(m.fileref, &mi);
				}
				ml->append(ml, &m);
			}
		}
	}
/*
  dmGetPath(path, dirref);
  if (!strcmp(path, "@:\\DEVICES\\"))
  {
    devinfonode *dev;
    for (dev=plWaveTableDevices; dev; dev=dev->next)
    {
      char hnd[9];
      strcpy(hnd, dev->handle);
      strupr(hnd);
      fsConvFileName12(m.name, hnd, ".DEV");
      if (fsMatchFileName12(m.name, mask))
      {
        m.dirref=dirref;
        m.fileref=mdbGetModuleReference(m.name, dev->dev.mem);
        if (m.fileref==0xFFFF)
          return 0;
        if (mdbGetModuleType(m.fileref)!=mtDEVw)
        {
          moduleinfostruct mi;
          mdbGetModuleInfo(mi, m.fileref);
          mi.flags1|=MDB_VIRTUAL;
          mi.channels=dev->dev.chan;
          strcpy(mi.modname, dev->name);
          mi.modtype=mtDEVw;
          mdbWriteModuleInfo(m.fileref, mi);
        }
        if (!mdbAppend(ml, m))
          return 0;
      }
    }
  }
*/
	return 1;
}

static int mcpSetDev(const char *path, struct moduleinfostruct *mi, FILE **fp)
{
	char name[9]; /* we never make names that are too long */
	_splitpath(path, 0, 0, name, 0);
	mcpSetDevice(name, 1);
	/*delay(1000); do we really need this ??? (doj)*/
	return 0;
}

static void mcpPrep(const char *path, struct moduleinfostruct *info, FILE **bp)
{
	mcpResetDevice();
/*	if (info->gen.moduleflags&MDB_BIGMODULE)           TODO
		mcpSetDevice(cfGetProfileString2(cfSoundSec, "sound", "bigmodules", ""), 0);  */
}


struct mdbreaddirregstruct mcpReadDirReg = {mcpReadDir};

struct interfacestruct mcpIntr = {mcpSetDev, 0, 0};

struct initcloseregstruct mcpDevReg={wavedevinit, wavedevclose};
struct preprocregstruct mcpPreprocess = {mcpPrep};
char *dllinfo = "initclose mcpDevReg; preprocess mcpPreprocess; readdirs mcpReadDirReg; interface mcpIntr";
struct linkinfostruct dllextinfo = {"mcpbase", "OpenCP Wavetable Devices System (c) 1994-04 Niklas Beisert, Tammo Hinrichs", DLLVERSION, 0};
