#ifndef __MP_H
#define __MP_H

#include "gmdinst.h"

#define MP_MAXCHANNELS 32

struct sampleinfo;

struct gmdtrack
{
	unsigned char *ptr;
	unsigned char *end;
};

struct gmdpattern
{
	char name[32];
	unsigned short patlen;
	unsigned short gtrack;
	unsigned short tracks[MP_MAXCHANNELS];
};


#define MOD_TICK0 1
#define MOD_EXPOFREQ 2
#define MOD_S3M 4
#define MOD_GUSVOL 8
#define MOD_EXPOPITCHENV 16
#define MOD_S3M30 32
#define MOD_MODPAN 0x10000

struct gmdmodule
{
	char name[32];
	char composer[32];
	unsigned long options;
	int channum;
	int instnum;
	int patnum;
	int ordnum;
	int endord;
	int loopord;
	int tracknum;
	int sampnum;
	int modsampnum;
	int envnum;
	struct gmdinstrument *instruments;
	struct gmdtrack *tracks;
	struct gmdenvelope *envelopes;
	struct sampleinfo *samples;
	struct gmdsample *modsamples;
	struct gmdpattern *patterns;
	char **message;
	unsigned short *orders;
};

struct globinfo
{
	unsigned char speed;
	unsigned char curtick;
	unsigned char tempo;
	unsigned char currow;
	unsigned short patlen;
	unsigned short curpat;
	unsigned short patnum;
	unsigned char globvol;
	unsigned char globvolslide;
};

struct chaninfo
{
	unsigned char ins;
	unsigned short smp;
	unsigned char note;
	unsigned char vol;
	unsigned char pan;
	unsigned char notehit;
	unsigned char volslide;
	unsigned char pitchslide;
	unsigned char panslide;
	unsigned char volfx;
	unsigned char pitchfx;
	unsigned char notefx;
	unsigned char fx;
};

extern void mpReset(struct gmdmodule *m);
extern void mpFree(struct gmdmodule *m);
extern int mpAllocInstruments(struct gmdmodule *m, int n);
extern int mpAllocSamples(struct gmdmodule *m, int n);
extern int mpAllocModSamples(struct gmdmodule *m, int n);
extern int mpAllocTracks(struct gmdmodule *m, int n);
extern int mpAllocPatterns(struct gmdmodule *m, int n);
extern int mpAllocEnvelopes(struct gmdmodule *m, int n);
extern int mpAllocOrders(struct gmdmodule *m, int n);
extern void mpOptimizePatLens(struct gmdmodule *m);
extern void mpReduceInstruments(struct gmdmodule *m);
extern void mpReduceMessage(struct gmdmodule *m);
extern int mpReduceSamples(struct gmdmodule *m);
extern int mpLoadSamples(struct gmdmodule *m);
extern void mpRemoveText(struct gmdmodule *m);

extern int mpLoadMOD(struct gmdmodule *, FILE *);
extern int mpLoadMODt(struct gmdmodule *, FILE *);
extern int mpLoadMODd(struct gmdmodule *, FILE *);
extern int mpLoadM15(struct gmdmodule *, FILE *);
extern int mpLoadM15t(struct gmdmodule *, FILE *);
extern int mpLoadM31(struct gmdmodule *, FILE *);
extern int mpLoadWOW(struct gmdmodule *, FILE *);
extern int mpLoadMTM(struct gmdmodule *, FILE *);
extern int mpLoadS3M(struct gmdmodule *, FILE *);
extern int mpLoadXM (struct gmdmodule *, FILE *);
extern int mpLoad669(struct gmdmodule *, FILE *);
extern int mpLoadULT(struct gmdmodule *, FILE *);
extern int mpLoadDMF(struct gmdmodule *, FILE *);
extern int mpLoadOKT(struct gmdmodule *, FILE *);
extern int mpLoadPTM(struct gmdmodule *, FILE *);
extern int mpLoadAMS(struct gmdmodule *, FILE *);
extern int mpLoadMDL(struct gmdmodule *, FILE *);
extern int mpLoadIT (struct gmdmodule *, FILE *);

extern char mpPlayModule(const struct gmdmodule *);
extern void mpStopModule(void);
extern void mpSetPosition(signed short pat, signed short row);
extern void mpGetPosition(uint16_t *pat, uint8_t *row);
extern int mpGetRealPos(void);
extern void mpGetChanInfo(unsigned char ch, struct chaninfo *ci);
extern unsigned short mpGetRealNote(unsigned char ch);
extern void mpGetGlobInfo(struct globinfo *gi);
extern char mpLooped(void );
extern void mpSetLoop(unsigned char s);
extern void mpLockPat(int st);
extern int mpGetChanSample(int ch, short *buf, int len, int rate, int opt);
extern void mpMute(int ch, int m);
extern void mpGetRealVolume(int ch, int *l, int *r);
extern int mpGetChanStatus(int ch);
extern int mpGetMute(int ch);

enum
{
	cmdTempo, cmdSpeed, cmdBreak, cmdGoto, cmdPatLoop, cmdPatDelay, cmdGlobVol, cmdGlobVolSlide, cmdSetChan, cmdFineSpeed
};

enum
{
	cmdVolSlideUp, cmdVolSlideDown, cmdRowVolSlideUp, cmdRowVolSlideDown,
	cmdPitchSlideUp, cmdPitchSlideDown, cmdPitchSlideToNote,
	cmdRowPitchSlideUp, cmdRowPitchSlideDown,
	cmdPanSlide, cmdRowPanSlide,
	cmdDelay,
	cmdVolVibrato, cmdVolVibratoSetWave, cmdTremor,
	cmdPitchVibrato, cmdPitchVibratoSetSpeed, cmdPitchVibratoFine,
	cmdPitchVibratoSetWave, cmdArpeggio,
	cmdNoteCut, cmdRetrig,
	cmdOffset,
	cmdPanSurround,
	cmdKeyOff,
	cmdSetEnvPos,

	cmdVolSlideUDMF, cmdVolSlideDDMF,
	cmdPanSlideLDMF, cmdPanSlideRDMF,
	cmdPitchSlideUDMF, cmdPitchSlideDDMF, cmdPitchSlideNDMF, cmdRowPitchSlideDMF,
	cmdVolVibratoSinDMF, cmdVolVibratoTrgDMF, cmdVolVibratoRecDMF,
	cmdPanVibratoSinDMF,
	cmdPitchVibratoSinDMF, cmdPitchVibratoTrgDMF, cmdPitchVibratoRecDMF,

	cmdPanDepth,
	cmdPanHeight,

	cmdChannelVol,

	cmdSpecial,
	cmdOffsetHigh,

	cmdOffsetEnd,
	cmdSetDir,
	cmdSetLoop,

	cmdPlayNote=0x80, cmdPlayIns=0x01, cmdPlayNte=0x02, cmdPlayVol=0x04, cmdPlayPan=0x08, cmdPlayDelay=0x10
};

enum
{
	cmdContVolSlide,
	cmdContRowVolSlide,
	cmdContMixVolSlide,
	cmdContMixVolSlideUp,
	cmdContMixVolSlideDown,
	cmdContMixPitchSlideUp,
	cmdContMixPitchSlideDown,
	cmdGlissOn,
	cmdGlissOff
};

enum
{
	fxGVSUp=1, fxGVSDown,
	fxVSUp=1, fxVSDown, fxVSUDMF, fxVSDDMF,
	fxPSUp=1, fxPSDown, fxPSToNote, fxPSUDMF, fxPSDDMF, fxPSNDMF,
	fxPnSRight=1, fxPnSLeft, fxPnSLDMF, fxPnSRDMF,
	fxVXVibrato=1, fxVXTremor,
	fxPXVibrato=1, fxPXArpeggio,
	fxPnXVibrato=1,
	fxNXNoteCut=1, fxNXRetrig,

	fxVolSlideUp=1, fxVolSlideDown,
	fxRowVolSlideUp, fxRowVolSlideDown,
	fxPitchSlideUp, fxPitchSlideDown, fxPitchSlideToNote,
	fxRowPitchSlideUp, fxRowPitchSlideDown,
	fxPanSlideRight, fxPanSlideLeft,
	fxVolVibrato, fxTremor,
	fxPitchVibrato, fxArpeggio,
	fxNoteCut, fxRetrig,
	fxOffset,
	fxDelay,
	fxPanVibrato
};

#endif
