/* OpenCP Module Player
 * copyright (c) '94-'98 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * GMIPlay Initialisiation (reads GUS patches etc)
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -kbwhenever Tammo Hinrichs <opencp@gmx.net>
 *    -changed path searching for ULTRASND.INI and patch files
 *    -corrected some obviously wrong allocations
 */

#include "config.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "types.h"
#include "gmiplay.h"
#include "binfile/binfpak.h"
#include "dev/mcp.h"
#include "boot/psetting.h"
#include "stuff/compat.h"

int midUseFFF=0;

unsigned char midInitFFF();

unsigned char midInit(const char *ipath)
{
	FILE *inifile;
	int i;
      	char path[PATH_MAX+1];
      	char path0[PATH_MAX+1];
	char path1[PATH_MAX+1];
	char *buf;
	char *bp;
	char *bp2;
	char type;
	long len;

	midUseFFF=cfGetProfileBool("midi", "usefff", 0, 0);
	if (midUseFFF)
		return midInitFFF();
	for (i=0; i<4; i++)
	{
		midInstrumentPaths[i]=calloc(256, 1);
		midInstrumentPaths[i][0]=0;
	}
	for (i=0; i<256; i++)
	{
		midInstrumentNames[i]=calloc(256, 1);
		midInstrumentNames[i][0]=0;
	}

	strcpy(path0,ipath);
	if (path0[strlen(path0)-1]!='/')
		strcat(path0, "/");
	strcpy(path1,path0);
	strcat(path1,"MIDI/");

	strcpy(midInstrumentPaths[1],path0);
	strcpy(midInstrumentPaths[0],path1);
	strcpy(midInstrumentPaths[2],cfDataDir);
	strcat(midInstrumentPaths[2],"MIDI/");

	strcat(path0, "ULTRASND.INI");
	strcat(path1, "ULTRASND.INI");

	if (!(inifile=fopen(path0, "r")))
	if (!(inifile=fopen(path1, "r")))
	if (!(inifile=pakfOpen("ULTRASND.INI")))
		return 0;

	fseek(inifile, 0, SEEK_END);
	len=ftell(inifile);
	fseek(inifile, 0, SEEK_SET);

	if (!(buf=calloc(len+1, 1)))
		return 0;
	fread(buf, len, 1, inifile);
  	buf[len]=0;
	fclose(inifile);

	*path=0;

	bp=buf;
	type=0;

	while (1)
	{
		int insnum;
		
		while (isspace(*bp))
			bp++;
		if (!*bp)
			break;

		if (*bp=='[')
		{
			if (!memicmp(bp, "[Melodic Bank 0]", 16))
				type=1;
			else
				if (!memicmp(bp, "[Drum Bank 0]", 13))
					type=2;
				else
					type=0;
		}

		if (!memicmp(bp, "PatchDir", 8))
		{
			while ((*bp!='=')&&*bp)
				bp++;
			if (*bp)
				bp++;
			while ((*bp==' ')||(*bp=='\t'))
				bp++;
			bp2=bp;
			while (!isspace(*bp2)&&*bp2)
				bp2++;
			memcpy(path, bp, bp2-bp);
			path[bp2-bp]=0;
			if (path[strlen(path)-1]!='/')
				strcat(path, "/");
		}
		
		if (!isdigit(*bp)||!type)
		{
			while (*bp&&(*bp!='\r')&&(*bp!='\n'))
				bp++;
			continue;
		}

		insnum=strtoul(bp, 0, 10)+((type==2)?128:0);
		while ((*bp!='=')&&*bp)
			bp++;
		if (*bp)
			bp++;
		while ((*bp==' ')||(*bp=='\t'))
			bp++;
		bp2=bp;
		while (!isspace(*bp2)&&*bp2)
			bp2++;

		if (insnum<256)
		{
			char *x=midInstrumentNames[insnum];
			strcpy(x+(bp2-bp), ".PAT");
			memcpy(x, bp, bp2-bp);
		}

		while (*bp&&(*bp!='\r')&&(*bp!='\n'))
			bp++;

	}
	free(buf);
	strcpy(midInstrumentPaths[3],path);

	return 1;
}

void midClose(void)
{
	int i;

	if (midUseFFF)
		closeFFF();
	else
	{
		for (i=0; i<4; i++)
			free(midInstrumentPaths[i]);
		for (i=0; i<256; i++)
			free(midInstrumentNames[i]);
	}
}

unsigned char midInitFFF(void)
{
	FILE *fff;
	const char *fn=cfGetProfileString("midi", "fff", "midi.fff");
      	char path[PATH_MAX+NAME_MAX+1];
	if(!(fff=fopen(fn, "r")))
	{
		if (!strchr(fn, '/'))
		{
			snprintf(path, sizeof(path), "%s%s", cfDataDir, fn);
			fff=fopen(fn, "r");
		}
	}
	if (!fff)
	{
		fprintf(stderr, "[FFF]: %s not found...\n", fn);
		return 0;
	}

	return loadFFF(fff);
}
