/* OpenCP Module Player
 * copyright (c) '94-'98 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * GMIPlay MID/RMI file loader
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -kbwhenever Tammo Hinrichs <opencp@gmx.net>
 *    -changed path searching for ULTRASND.INI and patch files
 *  -ryg_xmas   Fabian Giesen  <fabian@jdcs.su.nw.schule.de>
 *    -the .FFF hack, part I - untested, not integrated. wish you a nice
 *     time debugging it...
 *  -fd990122   Felix Domke    <tmbinc@gmx.net>
 *    -integrated the .FFF-hack, improved it, removed some silly bugs,
 *     but it's still not a loader... (and it's far away from that.)
 *     (some hours later: ok, some things are really loaded now.
 *      envelopes are ignored completely, maybe THIS is the error...
 *      you won't hear anything yet, sorry.)
 *     anyway: i am wondering how much changes we need to support all
 *     features of that fff :(
 *  -fd990124   Felix Domke    <tmbinc@gmx.net>
 *    -continued on the work. hmm, some sound will come out of your speakers
 *     right now, but it sounds just TERRIBLE. :)
 *  -ryg990125  Fabian Giesen  <fabian@jdcs.su.nw.schule.de>
 *    -corrected tmbs formulas by combining beisert's ones with some from
 *     GIPC (as Curtis Patzer, programmer of GIPC, would put it: a glorious
 *     hack). and, believe me, the songs start to sound like they should :)
 *    -added mu-law decoding table (nicer than your code, tmb)
 *    -now displays also instrument names (if program number in GM range
 *     this means <128)
 *  -kb990208  Tammo Hinrichs <kb@vmo.nettrade.de>
 *    -fixed some too obvious bugs
 *     ( for (i=...;..;..)
 *       {
 *         ...
 *         for (i=...;..;..) ... ;
 *         ...
 *       }; et al.)
 */

//              can anybody explain me in which ranges all the values are? ;)

/*
        do THIS in your cp.ini...

[midi]
  usefff=yes                ; ...or just "no" if you still want to hear midis
  dir=d:\utoplite\          ; where are your DATs located? :)
                            ; note: if you use fffs converted by gipc,
                            ; just write "dir=", because gipc includes
                            ; the path to the .dat-file.
  fff=d:\utoplite\utopi_li.fff

*/

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <unistd.h>
#include "types.h"
#include "boot/psetting.h"
#include "dev/mcp.h"
#include "gmiplay.h"
#include "stuff/compat.h"
#include "stuff/err.h"
#include "stuff/imsrtns.h"

#define DEBUG

/*
 * u-law (mu-law) decoding table (shamelessly stolen from SOX)
 */

static int ulaw_exp_table[] =
{
	-32124,-31100,-30076,-29052,-28028,-27004,-25980,-24956,
	-23932,-22908,-21884,-20860,-19836,-18812,-17788,-16764,
	-15996,-15484,-14972,-14460,-13948,-13436,-12924,-12412,
      	-11900,-11388,-10876,-10364, -9852, -9340, -8828, -8316,
	-7932, -7676, -7420, -7164, -6908, -6652, -6396, -6140,
	-5884, -5628, -5372, -5116, -4860, -4604, -4348, -4092,
	-3900, -3772, -3644, -3516, -3388, -3260, -3132, -3004,
	-2876, -2748, -2620, -2492, -2364, -2236, -2108, -1980,
	-1884, -1820, -1756, -1692, -1628, -1564, -1500, -1436,
	-1372, -1308, -1244, -1180, -1116, -1052,  -988,  -924,
	-876,  -844,  -812,  -780,  -748,  -716,  -684,  -652,
	-620,  -588,  -556,  -524,  -492,  -460,  -428,  -396,
	-372,  -356,  -340,  -324,  -308,  -292,  -276,  -260,
	-244,  -228,  -212,  -196,  -180,  -164,  -148,  -132,
	-120,  -112,  -104,   -96,   -88,   -80,   -72,   -64,
	-56,   -48,   -40,   -32,   -24,   -16,    -8,     0,
	32124, 31100, 30076, 29052, 28028, 27004, 25980, 24956,
	23932, 22908, 21884, 20860, 19836, 18812, 17788, 16764,
	15996, 15484, 14972, 14460, 13948, 13436, 12924, 12412,
	11900, 11388, 10876, 10364,  9852,  9340,  8828,  8316,
	7932,  7676,  7420,  7164,  6908,  6652,  6396,  6140,
	5884,  5628,  5372,  5116,  4860,  4604,  4348,  4092,
	3900,  3772,  3644,  3516,  3388,  3260,  3132,  3004,
	2876,  2748,  2620,  2492,  2364,  2236,  2108,  1980,
	1884,  1820,  1756,  1692,  1628,  1564,  1500,  1436,
	1372,  1308,  1244,  1180,  1116,  1052,   988,   924,
	876,   844,   812,   780,   748,   716,   684,   652,
	620,   588,   556,   524,   492,   460,   428,   396,
	372,   356,   340,   324,   308,   292,   276,   260,
	244,   228,   212,   196,   180,   164,   148,   132,
	120,   112,   104,    96,    88,    80,    72,    64,
	56,    48,    40,    32,    24,    16,     8,     0
};


int addpatchFFF(struct minstrument *, uint8_t, unsigned char, unsigned char, struct sampleinfo *, uint16_t *);

int loadpatchFFF(struct minstrument *, uint8_t, unsigned char*, struct sampleinfo **, uint16_t *);

char *midInstrumentPaths[4];
char *midInstrumentNames[256];


/* this is more nice, isn't it?
 *   -mw
 */
#include <netinet/in.h>
#define convshort htons
#define convlong htonl

/*static inline unsigned short convshort(uint8_t6_t a)
{
  return ((a&0xFF)<<8)|((a&0xFF00)>>8);
}
static inline unsigned long convlong(unsigned long a)
{
  return ((a&0xFF)<<24)|((a&0xFF00)<<8)|((a&0xFF0000)>>8)|((a&0xFF000000)>>24);
}
*/



static inline unsigned long readvlnum(unsigned char **ptr)
{
	unsigned long num=0;
	while (1)
	{
		num=(num<<7)|(*(*ptr)&0x7F);
		if (!(*(*ptr)++&0x80))
			break;
	}
	return num;
}

struct __attribute__((packed)) PATCHHEADER
{
	char header[12]; /* "GF1PATCH110" */
	char gravis_id[10];   /* "ID#000002" */
	char description[60];
	uint8_t instruments; /* 0 is bogus for 1 */
	uint8_t voices; /* should be 14*/
	uint8_t channels;
	uint16_t wave_forms;
	uint16_t master_volume;
	uint32_t data_size;
	char reserved[36];
};

struct __attribute__((packed)) INSTRUMENTDATA
{
	uint16_t instrument;
	char instrument_name[16];
	uint32_t instrument_size;
	uint8_t layers;
	char reserved[40];
};

struct __attribute__((packed)) LAYERDATA
{
	uint8_t layer_duplicate;
	uint8_t layer;
	uint32_t layer_size;
	uint8_t samples;
	char reserved[40];
};

struct __attribute__((packed)) PATCHDATA
{
	char wave_name[7];

	uint8_t fractions;
	uint32_t wave_size;
	uint32_t start_loop;
	uint32_t end_loop;

	uint16_t sample_rate;
	uint32_t low_frequency;
	uint32_t high_frequency;
	uint32_t root_frequency;
	uint16_t tune;

	uint8_t balance;

	uint8_t envelope_rate[6];
	uint8_t envelope_offset[6];

	uint8_t tremolo_sweep;
	uint8_t tremolo_rate;
	uint8_t tremolo_depth;

	uint8_t vibrato_sweep;
	uint8_t vibrato_rate;
	uint8_t vibrato_depth;

	/* bit 5 = Turn sustaining on. (Env. pts. 3)*/
	/* bit 6 = Enable envelopes - 1 */
	/* bit 7 = fast release */
	uint8_t modes;

	uint16_t scale_frequency;
	uint16_t scale_factor;    /* from 0 to 2048 or 0 to 2 */

	char reserved[36];
};

static uint32_t pocttab[16]={2044, 4088, 8176, 16352, 32703, 65406, 130813, 261626, 523251, 1046502, 2093005, 4186009, 8372018, 16744036, 33488072, 66976145};
static uint16_t pnotetab[12]={32768, 34716, 36781, 38968, 41285, 43740, 46341, 49097, 52016, 55109, 58386, 61858};
static uint16_t pfinetab[16]={32768, 32887, 33005, 33125, 33245, 33365, 33486, 33607, 33728, 33850, 33973, 34095, 34219, 34343, 34467, 34591};
static uint16_t pxfinetab[16]={32768, 32775, 32783, 32790, 32798, 32805, 32812, 32820, 32827, 32835, 32842, 32849, 32857, 32864, 32872, 32879};

static int16_t getnote(uint32_t frq)  /* frq=freq*1000, res=(oct*12+note)*256 (and maybe +fine*16+xfine) */
{
	uint16_t x;
	int i;
	for (i=0; i<15; i++)
		if (pocttab[i+1]>frq)
			break;
	x=(i-1)*12*256;
	frq=umuldiv(frq, 32768, pocttab[i]);
	for (i=0; i<11; i++)
		if (pnotetab[i+1]>frq)
			break;
	x+=i*256;
	frq=umuldiv(frq, 32768, pnotetab[i]);
	for (i=0; i<15; i++)
		if (pfinetab[i+1]>frq)
			break;
	x+=i*16;
	frq=umuldiv(frq, 32768, pfinetab[i]);
	for (i=0; i<15; i++)
		if (pxfinetab[i+1]>frq)
			break;
	return x+i;
}

static uint32_t getfreq(uint16_t note)
{
	int freq=pocttab[note/256/12+1];
	freq=umuldiv(freq, pnotetab[(note/256)%12], 32768);
	freq=umuldiv(freq, pfinetab[(note/16)&0xF], 32768);
	return(umuldiv(freq, pxfinetab[(note)&0xF], 32768));  /* (x-)finetuning not VERY much tested. */
}

/*
 * the .FFF-hack (loader? not really) begins *here*
 * (YES, i'm now using binfile-types, increases portability [as if that had
 *  any use... {nested braces rule! :) }])
 */

#ifdef __WATCOMC__
#pragma pack (push, 1)  // no structure alignment (tmb always forgets those)
#endif

struct __attribute__((packed)) FFF_CHUNK_HEADER
{
	char                        id[4];
	uint32_t                    size;
};

union __attribute__((packed)) FFF_ID
{
	struct
	{
		uint16_t            maj_id,
				    min_id;
	};

	uint32_t                    id;
	void                       *ptr;
};

struct __attribute__((packed)) FFF_ENVELOPE_POINT
{
	uint16_t                    next;

	union
	{
	       	uint16_t            rate;
		uint16_t            time;
	};
};

struct __attribute__((packed)) FFF_LFO
{
	uint16_t                    freq;
	int16_t                     depth;
	int16_t                     sweep;
	uint8_t                     shape;         /* no ryg, a byte is NOT an uint8 ;) (fd)... it is now (sss) */
	uint8_t                     delay;         /* a normal rounding error, nothing more :) (ryg) */
};

struct __attribute__((packed)) FFF_ENVELOPE_RECORD
{
	int16_t                     nattack;
	int16_t                     nrelease;
	uint16_t                    sustain_offset;
	uint16_t                    sustain_rate;
	uint16_t                    release_rate;
	uint8_t                     hirange;
	uint8_t                     d0;
	struct FFF_ENVELOPE_POINT  *attack_points;
	struct FFF_ENVELOPE_POINT  *release_points;
};

struct __attribute__((packed)) FFF_ENVP_CHUNK
{
	union FFF_ID                id;
	uint8_t                     num_envelopes;
	uint8_t                     flags; /* bit0=retrigger, rest unused/unimportant */
	uint8_t                     mode;
	uint8_t                     indtype;
	struct FFF_ENVELOPE_RECORD *records;
};

struct __attribute__((packed)) FFF_PROG_CHUNK
{
	union FFF_ID                id;
	union FFF_ID                version;
};

struct __attribute__((packed)) FFF_WAVE_CHUNK
{
	union FFF_ID                id;
	uint32_t                    size;
	uint32_t                    start;
	uint32_t                    loopstart;
	uint32_t                    loopend;
	uint32_t                    m_start;
	uint32_t                    sample_ratio;
	uint8_t                     attenuation;
	uint8_t                     low_note;
	uint8_t                     high_note;
	uint8_t                     format;
	uint8_t                     m_format;
	union FFF_ID                data_id;
};

struct __attribute__((packed)) FFF_LAYR_CHUNK
{
	union FFF_ID           id;
	uint8_t                nwaves;
	uint8_t                flags;
	uint8_t                high_range;
	uint8_t                low_range;
	uint8_t                pan;
	uint8_t                pan_freq_scale;
	struct FFF_LFO         tremolo;
	struct FFF_LFO         vibrato;
	uint8_t                velocity_mode;
	uint8_t                attenuation;
	int16_t                freq_scale;
	uint8_t                freq_center;
	uint8_t                layer_event;
	union FFF_ID           penv;
	union FFF_ID           venv;
	struct FFF_WAVE_CHUNK *waves;
};

struct __attribute__((packed)) FFF_PTCH_CHUNK
{
	union FFF_ID           id;
	int16_t                nlayers;
	uint8_t                layer_mode;
	uint8_t                excl_mode;
	int16_t                excl_group;
	uint8_t                effect1;
	uint8_t                effect1_depth;
	uint8_t                effect2;
	uint8_t                effect2_depth;
	uint8_t                bank;
	uint8_t                program;
	struct FFF_LAYR_CHUNK *iw_layer;
};

struct __attribute__((packed)) FFF_DATA_CHUNK
{
	union FFF_ID            id;
	char                    filename[256];
};

#ifdef __WATCOMC__
#pragma pack (pop)  // reset structure packing
#endif

struct FFF_ENVP_LIST
{
	struct FFF_ENVP_CHUNK      *chunk;
	struct FFF_ENVP_LIST       *next;
};

struct FFF_PTCH_LIST
{
	struct FFF_PTCH_CHUNK      *chunk;
	struct FFF_PTCH_LIST       *next;
};

struct FFF_DATA_LIST
{
	struct FFF_DATA_CHUNK      *chunk;
	struct FFF_DATA_LIST       *next;
};


struct FFF_ENVP_LIST *envp_list=0;
struct FFF_PTCH_LIST *ptch_list=0;
struct FFF_DATA_LIST *data_list=0;

void *getENVP(uint32_t id)
{
	struct FFF_ENVP_LIST *l=envp_list;

	while (l)
	{
		if (l->chunk->id.id==id)
			return l->chunk;
		l=l->next;
	};

	return 0;
};

void *getDATA(uint32_t id)
{
	struct FFF_DATA_LIST *l=data_list;

	while (l)
	{
		if (l->chunk->id.id==id)
			return l->chunk;
		l=l->next;
	};

	return 0;
};

int loadFFF(FILE *file)
{
	struct FFF_CHUNK_HEADER  hd;
	int                      i, j, matched;
      	struct FFF_PTCH_LIST *l;

	/* pass one: read the .FFF file */
	if (fread(&hd, sizeof(hd), 1, file)!=1)
		return 0;

      	if (strncmp(hd.id, "FFFF", 4))
		return 0;

	if (fread(&hd, sizeof(hd), 1, file)!=1)
		return 0;

	while (1) /*(strncmp(hd.id, "FFFF", 4)) ryg, i don't understand your code... (fd)  me too :) (ryg) */
	{
		matched=0;

#ifdef DEBUG
	    	fprintf(stderr, "[FFF] got %c%c%c%c-chunk\n", hd.id[0], hd.id[1], hd.id[2], hd.id[3]);
#endif

		if (!strncmp(hd.id, "ENVP", 4))
		{
			struct FFF_ENVP_LIST       *l;
			struct FFF_ENVELOPE_RECORD *lcr;

			matched=1;

			l=calloc(sizeof(struct FFF_ENVP_LIST), 1);
			l->chunk=calloc(sizeof(struct FFF_ENVP_CHUNK), 1);
			l->next=envp_list;
			envp_list=l;

			if (fread(&l->chunk->id.id,          sizeof(uint32_t), 1, file)!=1) return 0;
			if (!fread(&l->chunk->num_envelopes, sizeof(uint8_t),  1, file)!=1) return 0;
			if (!fread(&l->chunk->flags,         sizeof(uint8_t),  1, file)!=1) return 0;
			if (!fread(&l->chunk->mode,          sizeof(uint8_t),  1, file)!=1) return 0;
			if (!fread(&l->chunk->indtype,       sizeof(uint8_t),  1, file)!=1) return 0;

			l->chunk->records=calloc(sizeof(struct FFF_ENVELOPE_RECORD), l->chunk->num_envelopes);
			lcr=l->chunk->records;

			for (i=0; i<l->chunk->num_envelopes; i++)
			{
				if (fread(&lcr[i].nattack,        sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lcr[i].nrelease,       sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lcr[i].sustain_offset, sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lcr[i].sustain_rate,   sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lcr[i].release_rate,   sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lcr[i].hirange,        sizeof(uint8_t),  1, file)!=1) return 0;
				if (fread(&lcr[i].d0,             sizeof(uint8_t),  1, file)!=1) return 0;

				if (lcr[i].nattack)
					lcr[i].attack_points=calloc(sizeof(struct FFF_ENVELOPE_POINT), lcr[i].nattack);
				else
					lcr[i].attack_points=0;

				if (lcr[i].nrelease)
					lcr[i].release_points=calloc(sizeof(struct FFF_ENVELOPE_POINT), lcr[i].nrelease);
				else
					lcr[i].release_points=0;
				if (lcr[i].nattack)
					if (fread(lcr[i].attack_points, sizeof(struct FFF_ENVELOPE_POINT)*lcr[i].nattack, 1, file)!=1)
						return 0;
				if (lcr[i].nrelease)
					if (fread(lcr[i].release_points, sizeof(struct FFF_ENVELOPE_POINT)*lcr[i].nrelease, 1, file)!=1)
						return 0;
			}
		}

		if (!strncmp(hd.id, "PROG", 4))
		{
			uint32_t dummy[2];

			matched=1;

			fread(dummy, sizeof(dummy), 1, file);
#ifdef DEBUG
			fprintf(stderr, "[FFF] skipped program\n");
#endif
		}

		if (!strncmp(hd.id, "PTCH", 4))
		{
			struct FFF_PTCH_LIST  *l;
			struct FFF_PTCH_CHUNK *c;
			struct FFF_LAYR_CHUNK *lr;
			uint32_t dummy;

#ifdef DEBUG
			fprintf(stderr, "[FFF] loading patch\n");
#endif

			matched=1;

			l=calloc(sizeof(struct FFF_PTCH_LIST), 1);
			l->chunk=calloc(sizeof(struct FFF_PTCH_CHUNK), 1);
			l->next=ptch_list;
			ptch_list=l;
			c=l->chunk;

			if (fread(&c->id.id, sizeof(uint32_t), 1, file)!=1) return 0;
			if (fread(&c->nlayers, sizeof(uint16_t), 1, file)!=1) return 0;
			if (fread(&c->layer_mode, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->excl_mode, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->excl_group, sizeof(uint16_t), 1, file)!=1) return 0;
			if (fread(&c->effect1, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->effect1_depth, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->effect2, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->effect2_depth, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->bank, sizeof(uint8_t), 1, file)!=1) return 0;
			if (fread(&c->program, sizeof(uint8_t), 1, file)!=1) return 0;

			/* actually, iw_layer isn't used on disk, but it's there (fd) yup (ryg) */
			fread(&dummy, sizeof(uint32_t), 1, file);

#ifdef DEBUG
			fprintf(stderr, "[FFF] %d:%d loading, %d layers.\n", c->bank, c->program, c->nlayers);
#endif

			c->iw_layer=calloc(sizeof(struct FFF_LAYR_CHUNK), c->nlayers);

			for (i=0; i<c->nlayers; i++)
			{
				struct FFF_CHUNK_HEADER  lrhd;
				struct FFF_WAVE_CHUNK   *wv;
				uint32_t dummy;

				lr=&c->iw_layer[i];

				if (fread(&lrhd, sizeof(lrhd), 1, file)!=1) return 0;

				/* sachmal ryg, wielange warste schon auf als du DAS hier gecoded hast? merke:
				 * strcmp==0 means "gleich"! ;) (fd) 17 stunden, wieso? (ryg)
				 */
				if (strncmp(lrhd.id, "LAYR", 4))
				{
					printf("[FFF] non-LAYR chunk in PTCH (malformed FFF file)\n");
					printf("[FFF] (found %c%c%c%c-chunk)\n", lrhd.id[0], lrhd.id[1], lrhd.id[2], lrhd.id[3]);
					return 0;
				}

				if (fread(&lr->id.id, sizeof(uint32_t), 1, file)!=1) return 0;
				if (fread(&lr->nwaves, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->flags, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->high_range, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->low_range, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->pan, sizeof(uint8_t), 1, file)!=1) return 0;
                                if (fread(&lr->pan_freq_scale, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->tremolo, sizeof(struct FFF_LFO), 1, file)!=1) return 0;
				if (fread(&lr->vibrato, sizeof(struct FFF_LFO), 1, file)!=1) return 0;
				if (fread(&lr->velocity_mode, sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->attenuation,  sizeof(uint8_t), 1, file)!=1) return 0;
				if (fread(&lr->freq_scale, sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lr->freq_center, sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lr->layer_event, sizeof(uint16_t), 1, file)!=1) return 0;
				if (fread(&lr->penv.id, sizeof(uint32_t), 1, file)!=1) return 0;
				if (fread(&lr->venv.id, sizeof(uint32_t), 1, file)!=1) return 0;

				fread(&dummy, sizeof(uint32_t), 1, file);

				lr->waves=calloc(sizeof(struct FFF_WAVE_CHUNK), lr->nwaves);

#ifdef DEBUG
				fprintf(stderr, "[FFF] ... %d waves.\n", lr->nwaves);
#endif

				for (j=0; j<lr->nwaves; j++)
				{
					struct FFF_CHUNK_HEADER  wvhd;

					wv=&lr->waves[j];

					if (fread(&wvhd, sizeof(wvhd), 1, file)!=1) return 0;

					if (strncmp(wvhd.id, "WAVE", 4))
					{
						fprintf(stderr, "[FFF] non-WAVE chunk in LAYR (malformed FFF file)\n");
						fprintf(stderr, "[FFF] (found %c%c%c%c-chunk)\n", wvhd.id[0], wvhd.id[1], wvhd.id[2], wvhd.id[3]);
						return 0;
					};

					if (fread(&wv->id.id, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->size, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->start, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->loopstart, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->loopend, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->m_start, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->sample_ratio, sizeof(uint32_t), 1, file)!=1) return 0;
					if (fread(&wv->attenuation, sizeof(uint8_t), 1, file)!=1) return 0;
					if (fread(&wv->low_note, sizeof(uint8_t), 1, file)!=1) return 0;
					if (fread(&wv->high_note, sizeof(uint8_t), 1, file)!=1) return 0;
					if (fread(&wv->format, sizeof(uint8_t), 1, file)!=1) return 0;
					if (fread(&wv->m_format, sizeof(uint8_t), 1, file)!=1) return 0;
					if (fread(&wv->data_id.id, sizeof(uint32_t), 1, file)!=1) return 0;
					fprintf(stderr, "wave %d loaded.. (%d bytes, playable from %x to %x)\n", i, wv->size, wv->low_note, wv->high_note);
				}
			}
		}

		if (!strncmp(hd.id, "DATA", 4))
		{
			struct FFF_DATA_LIST       *l;
#ifdef DEBUG
			fprintf(stderr, "[FFF] read DATA chunk\n");
#endif

			matched=1;

			l=calloc(sizeof(struct FFF_DATA_LIST), 1);
			l->chunk=calloc(sizeof(struct FFF_DATA_CHUNK), 1);
			l->next=data_list;
			data_list=l;
			if (fread(&l->chunk->id.id, sizeof(uint32_t), 1, file)!=1) return 0;
			if (fread(&l->chunk->filename[0], hd.size-4, 1, file)!=1) return 0;
		}

		if (!matched)
		{
			fseek(file, hd.size, SEEK_CUR);
#ifdef DEBUG
			fprintf(stderr, "[FFF] skipped %c%c%c%c-chunk\n", hd.id[0], hd.id[1], hd.id[2], hd.id[3]);
#endif
		}

		if (fread(&hd, sizeof(hd), 1, file)!=1) break;
	}

	/* pass 2: convert those IDs to pointers (yeeah) */

      	l=ptch_list;

	while (l)
	{
		for (i=0; i<l->chunk->nlayers; i++)
		{
			struct FFF_LAYR_CHUNK *lc;

			lc=&l->chunk->iw_layer[i];

			if (lc->penv.id)
			{
				lc->penv.ptr=getENVP(lc->penv.id);
				if (! lc->penv.ptr )
				{
					fprintf (stderr, "penvelop id %x not found.\n", lc->penv.id);
					return 0;
				}
			}

			if (lc->venv.id)
			{
				lc->venv.ptr=getENVP(lc->venv.id);
				if (! lc->venv.ptr )
				{
					fprintf (stderr, "venvelop id %x not found.\n", lc->venv.id);
					return 0;
				}
			}

			for (j=0; j<lc->nwaves; j++)
			{
				lc->waves[j].data_id.ptr=getDATA(lc->waves[j].data_id.id);
				if (! lc->waves[j].data_id.ptr )
				{
					fprintf (stderr, "wavedata file #%x not found...\n", lc->waves[j].data_id.id);
					return 0;
				}
			}
		}

		l=l->next;
	}

	fprintf(stderr, "done with that loading. be happy.\n");

	return 1;
};

/*
 * end of .FFF-hack
 */

static int loadsample(int                 file,
                      struct minstrument *ins,
                      uint8_t             j,
                      uint8_t             vox,
                      int                 setnote, /* local boolean */
                      uint8_t             sampnum,
                      uint8_t            *sampused,
                      struct sampleinfo  *sip,
                      uint16_t           *samplenum)
{
	struct msample *s=&ins->samples[j];
	int bit16; /* local boolean */
	struct PATCHDATA sh;
	int q;
	uint8_t *smpp;

	read(file, &sh, sizeof(sh));

	bit16=!!(sh.modes&1);
	if (bit16)
	{
		sh.wave_size>>=1;
		sh.start_loop>>=1;
		sh.end_loop>>=1;
	}

	if (setnote)
	{
		uint8_t lownote,highnote;
		int i;

		lownote=(getnote(sh.low_frequency)+0x80)>>8;
		highnote=(getnote(sh.high_frequency)+0x80)>>8;
		for (i=lownote; i<highnote; i++)
			if (sampused[i>>3]&(1<<(i&7)))
				break;
		if (i==highnote)
		{
			fprintf(stderr, "..... i==highnote\n");
			lseek(file, sh.wave_size<<bit16, SEEK_CUR);
			return 1;
		}
		memset(ins->note+lownote, j, highnote-lownote);
	}

	memcpy(s->name, sh.wave_name, 7);
	s->name[7]=0;
	s->sampnum=sampnum;
	s->handle=-1;
	s->normnote=getnote(sh.root_frequency);
	if ((s->normnote&0xFF)>=0xFE)
		s->normnote=(s->normnote+2)&~0xFF;
	if ((s->normnote&0xFF)<=0x02)
		s->normnote=s->normnote&~0xFF;
	sip->length=sh.wave_size;
	sip->loopstart=sh.start_loop;
	sip->loopend=sh.end_loop;
	sip->samprate=sh.sample_rate;
	sip->type=((sh.modes&4)?(mcpSampLoop|((sh.modes&8)?mcpSampBiDi:0)):0)|(bit16?mcpSamp16Bit:0)|((sh.modes&2)?mcpSampUnsigned:0);
/*	fprintf(stderr, "env: "); */
	for (q=0; q<6; q++)
	{
/*		s->volrte[q]=((sh.envelope_rate[q]&63)<<(9-3*(sh.envelope_rate[q]>>6)))*1220*16/(64*vox)*2/3; */
		s->volrte[q]=((int)(((((int)sh.envelope_rate[q])&63)*11025))>>((int)3*((int)(sh.envelope_rate[q]>>6))))*14/(int)vox;
		s->volpos[q]=((int)sh.envelope_offset[q])<<8;
/*		if (q<s->end) fprintf(stderr, "%d:%d ", s->volrte[q], s->volpos[q]); */
	}
/*	fprintf(stderr, "\n"); */

	s->end=(sh.modes&128)?3:6;
	s->sustain=(sh.modes&32)?3:7;
	s->tremswp=((int)sh.tremolo_sweep)*(int)64/45;
	s->vibswp=((int)sh.vibrato_sweep)*(int)64/45;
	s->tremdep=((int)sh.tremolo_depth*4*(int)256)/255   /2;
	s->vibdep=((int)sh.vibrato_depth)*12*(int)256/255   /4;
	s->tremrte=(((int)sh.tremolo_rate)*7+15)*65536/(300*64);
	s->vibrte=(((int)sh.vibrato_rate)*7+15)*65536/(300*64);

/*	fprintf(stderr, "   -> %d %d %d, %d %d %d\n", s->tremswp, s->tremrte, s->tremdep,
	                                              s->vibswp, s->vibrte, s->vibdep);
 */

	if (sh.scale_factor<=2)
		s->sclfac=((int)sh.scale_factor)<<8;
	else
		s->sclfac=sh.scale_factor>>2;
	s->sclbas=sh.scale_frequency;

	if (!(smpp=calloc(sip->length<<bit16, 1)))
		return errAllocMem;
	if (read(file, smpp, sip->length<<bit16)!=(sip->length<<bit16))
		fprintf(stderr, "short read...\n");
	sip->ptr=smpp;
	s->handle=(*samplenum)++;
	fprintf(stderr, "(1)Loaded sample %x ptr=%p samples=%ld (16bit=%d)\n", s->handle, sip->ptr, sip->length, bit16);
	return errOk;
}


int loadpatch(struct minstrument *ins,
              uint8_t             program,
              uint8_t            *sampused,
              struct sampleinfo **smps,
              uint16_t           *samplenum)
{
      	int file;
	char path[PATH_MAX+NAME_MAX];
	int p;
      	struct PATCHHEADER ph;
	struct INSTRUMENTDATA ih;
      	struct LAYERDATA lh;
	int i;
	int cursamp=0;
	uint8_t lowest;
	if (midUseFFF)
		return loadpatchFFF(ins, program, sampused, smps, samplenum);

	ins->sampnum=0;
	*ins->name=0;


	for (p=0; p<4; p++)
	{
		snprintf(path, sizeof(path), "%s%s", midInstrumentPaths[p], midInstrumentNames[program]);
		if ((file=open(path, O_RDONLY))>0)
			break;
	}
	if (file<0)
	{
		fprintf(stderr, "Error: Patch file %s not found!\n", midInstrumentNames[program]);
		return errFileMiss;
	}

	fprintf(stderr, "loadpatch loading file %s\n", path);

	read(file, &ph, sizeof(ph));

	if (memcmp(ph.header, "GF1PATCH110", 12))
		return errFormStruc;
	if (ph.instruments<1)
		return errFormStruc;

	read(file, &ih, sizeof(ih));

	if (ih.layers<1)
		return errFormStruc;
	strcpy(ins->name, ih.instrument_name);
	ins->name[16]=0;
	if (!*ins->name)
	{
		char name[NAME_MAX+1];
		_splitpath(midInstrumentNames[program], 0, 0, name, 0);
		snprintf(ins->name, sizeof(ins->name), "%s", name);
	}

	read(file, &lh, sizeof(lh));

	if (!(ins->samples=calloc(sizeof(struct msample), lh.samples)))
	{
		close(file);
		return errFormStruc;
	}
	if (!(*smps=calloc(sizeof(struct sampleinfo), lh.samples)))
	{
		close(file);
		return errFormStruc;
	}
	ins->sampnum=lh.samples;
	memset(*smps, 0, lh.samples*sizeof(**smps));

	memset(ins->note, 0xFF, 0x80);

	for (i=0; i<ins->sampnum; i++)
	{
		int8_t st=loadsample(file, ins, cursamp, ph.voices, 1, i, sampused, (*smps)+cursamp, samplenum);
		if (st<errOk)
		{
			close(file);
			return st;
		}
		if (st!=1)
			cursamp++;
	}
	ins->sampnum=cursamp;

	lowest=0xFF;
	for (i=0; i<0x80; i++)
		if (ins->note[i]!=0xFF)
		{
			lowest=ins->note[i];
			break;
		}

	for (i=0; i<0x80; i++)
		if (ins->note[i]!=0xFF)
			lowest=ins->note[i];
		else
			ins->note[i]=lowest;

	close(file);

	return errOk;
}


int addpatch( struct minstrument *ins,
              uint8_t             program,
              uint8_t             sn,
              uint8_t             sampnum,
              struct sampleinfo  *sip,
              uint16_t           *samplenum)
{
	struct msample *s;
	int file;
	char path[PATH_MAX+NAME_MAX];
	int p, st;
	struct PATCHHEADER ph;
	struct INSTRUMENTDATA ih;
	struct LAYERDATA lh;

	if (midUseFFF)
		return addpatchFFF(ins, program, sn, sampnum, sip, samplenum);

	s=&ins->samples[sn];

	for (p=0; p<4; p++)
	{
		snprintf(path, sizeof(path), "%s%s", midInstrumentPaths[p], midInstrumentNames[program]);
		file=open(path, O_RDONLY);
		if (file>=0)
			break;
	}
	if (file<0)
	{
		fprintf(stderr, "Error: Patch file %s not found!\n", midInstrumentNames[program]);
		return errFileMiss;
	}

	fprintf(stderr, "Adding patch from %s\n", path);

      	read(file, &ph, sizeof(ph));

	if (memcmp(ph.header, "GF1PATCH110", 12))
	{
		fprintf(stderr, "Invalid version...\n");
		close(file);
		return errFormStruc;
	}
	if (ph.instruments>1)
	{
		fprintf(stderr, "ph.instruments>1...\n");
		close(file);
		return errFormStruc;
	}
	read(file, &ih, sizeof(ih));

	if (ih.layers<1)
	{
		int q;
		uint8_t *smpp;
/*		struct msample s=ins->samples[sn];  overkill */

		strcpy(s->name, "no sample");
		s->handle=-1;
		s->sampnum=sampnum;
		s->normnote=getnote(440000);
		sip->length=1;
		sip->loopstart=0;
		sip->loopend=0;
		sip->samprate=44100;
		sip->type=0;
		for (q=0; q<6; q++)
		{
			s->volpos[q]=0;
			s->volrte[q]=0;
		}
		s->end=1;
		s->sustain=-1;
		s->vibdep=0;
		s->vibrte=0;
		s->vibswp=0;
		s->tremdep=0;
		s->tremrte=0;
		s->tremswp=0;
		s->sclfac=256;
		s->sclbas=60;

		if (!(smpp=malloc(sizeof(uint8_t))))
		{
			close(file);
			return errAllocMem;
		}
		*smpp=0;
		sip->ptr=smpp;
		s->handle=(*samplenum)++;
		fprintf(stderr, "(2)Loaded sample %x\n", s->handle);

		close(file);
		return 0;
	}

	read(file, &lh, sizeof(lh));

	if (lh.samples!=1)
	{
		fprintf(stderr, "lh.samples!=1\n");
		close(file);
		return errFormStruc;
	}

	if ((st=loadsample(file, ins, sn, ph.voices, 0, sampnum, 0, sip, samplenum)))
		return st;

	strcpy(s->name, ih.instrument_name);
	s->name[16]=0;
	if (!*s->name)
	{
		char name[NAME_MAX+1];
		_splitpath(midInstrumentNames[program], 0, 0, name, 0);
		snprintf(s->name, sizeof(s->name), "%s", name);
	}

	close(file);

	return errOk;
}


char midLoadMidi( struct midifile *m,
                  FILE            *file,
                  uint32_t         opt,
                  const char      *gpp)
{
	mid_free(m);
	uint8_t drumch2;
	uint32_t len;

	uint16_t trknum;
	uint16_t mtype;

	uint32_t dummy;
	int i, j;
	int samplenum;

	uint8_t (*sampused)[16];
	uint8_t instused[0x81];
	uint8_t chaninst[16];

	struct sampleinfo **smps;
	uint16_t inst;

	m->opt=opt;

	drumch2=(m->opt&MID_DRUMCH16)?15:16;

	while (1)
	{
		uint32_t type;

	    	fread(&type, sizeof(uint32_t), 1, file);

		if (type==0x46464952)
		{
			fread(&dummy, sizeof(uint32_t), 1, file);
			fread(&dummy, sizeof(uint32_t), 1, file);

			if (dummy!=0x44494D52)
				return errFormStruc;
			while (1)
			{
				fread(&dummy, sizeof(uint32_t), 1, file);
				if (dummy==0x61746164)
					break;
				fread(&dummy, sizeof(uint32_t), 1, file);
				fseek(file, dummy, SEEK_CUR);
			}
		fread(&dummy, sizeof(uint32_t), 1, file);
		continue;
		}
		if (type==*(uint32_t *)"MThd")
			break;
		fread(&dummy, sizeof(uint32_t), 1, file);
		fseek(file, convlong(dummy), SEEK_CUR);
	}

	fread(&dummy, sizeof(uint32_t), 1, file);
	len=convlong(dummy);
	if (len<6)
		return errFormStruc;

	fread(&mtype, sizeof(uint16_t), 1, file);
	mtype=convshort(mtype);
	fread(&trknum, sizeof(uint16_t), 1, file);
	trknum=convshort(trknum);
	fread(&m->tempo, sizeof(uint16_t), 1, file);
	m->tempo=convshort(m->tempo);
	fseek(file, len-6, SEEK_CUR);

	if (mtype>=3)
		return errFormSupp;
	if ((mtype==1)&&(trknum>64))
		return errFormSupp;

	m->tracknum=(mtype==1)?trknum:1;

	if (!(m->tracks=calloc(sizeof(struct miditrack), m->tracknum)))
		return errAllocMem;

	for (i=0; i<m->tracknum; i++)
	{
		m->tracks[i].trk=0;
		m->tracks[i].trkend=0;
	}

	for (i=0; i<trknum; i++)
	{
		while (1)
		{
			uint32_t type;
			if (fread(&type, sizeof(uint32_t), 1, file)!=1)
				return errFormStruc;
			fread(&len, sizeof(uint32_t), 1, file);
			len=convlong(len);
			if (type==*(unsigned long *)"MTrk")
				break;
			fseek(file, len, SEEK_CUR);
		}

		if (mtype!=2)
		{
			if (!(m->tracks[i].trk=calloc(len, 1)))
				return errAllocMem;
			m->tracks[i].trkend=m->tracks[i].trk+len;
			fread(m->tracks[i].trk, len, 1, file);
		} else {
			uint32_t oldlen=m->tracks[0].trkend-m->tracks[0].trk;
			uint8_t *n=(uint8_t *)realloc(m->tracks[0].trk, oldlen+len);
			if (!n)
				return errAllocMem;
			m->tracks[0].trk=n;
			m->tracks[0].trkend=n+oldlen+len;
			fread(m->tracks[0].trk+oldlen, len, 1, file);
		}
	}

	if (!(sampused=calloc(0x81, 16)))
		return errAllocMem;

	memset(instused, 0, 0x81);
	memset(chaninst, 0, 16);
	chaninst[9]=0x80;
	if (drumch2<16)
		chaninst[drumch2]=0x80;

	m->ticknum=0;
	for (i=0; i<m->tracknum; i++)
	{
		uint8_t *trkptr=m->tracks[i].trk;
		uint8_t status=0;
		uint32_t trackticks=0;
		while (trkptr<m->tracks[i].trkend)
		{
			trackticks+=readvlnum(&trkptr);
			if (*trkptr&0x80)
				status=*trkptr++;
			if ((status==0xFF)||(status==0xF0)||(status==0xF7))
			{
				if (status==0xFF)
					trkptr++;
				trkptr+=readvlnum(&trkptr);
			} else
				switch (status&0xF0)
				{
					case 0x90:
						if (trkptr[1])
						{
							sampused[chaninst[status&0xF]][trkptr[0]>>3]|=1<<(trkptr[0]&7);
							instused[chaninst[status&0xF]]=1;
						}
						trkptr+=2;
						break;
					case 0x80: case 0xA0: case 0xB0: case 0xE0:
						trkptr+=2;
						break;
					case 0xD0:
						trkptr++;
						break;
					case 0xC0:
						if (((status&0xF)!=9)&&((status&0xF)!=drumch2))
						{
							chaninst[status&0xF]=trkptr[0];
							/* shit! */
							memset(sampused[trkptr[0]], 0xFF, 16);
							instused[trkptr[0]]=1;
						}
						trkptr++;
						break;
				}
		}
		if (m->ticknum<trackticks)
			m->ticknum=trackticks;
	}
	if (!m->ticknum)
		return errFormStruc;

	m->instnum=0;
	for (i=0; i<0x81; i++)
		if (instused[i])
			m->instnum++;

	if (!m->instnum)
	{
		instused[0]=1;
		memset(sampused[0], 0xFF, 16);
		m->instnum++;
	}

	if (!(smps=calloc(sizeof(struct sampleinfo *), m->instnum)))
		return errAllocMem;
	if (!(m->instruments=calloc(sizeof(struct minstrument), m->instnum)))
	{
		free(smps);
		return errAllocMem;
	}

	for (i=0; i<m->instnum; i++)
	{
		m->instruments[i].sampnum=0;
		m->instruments[i].samples=0;
		smps[i]=0;
	}

	if (!midInit(gpp))
	{
		free(smps);
		midClose();
		return errFileMiss;
	}

	m->sampnum=0;
	memset(m->instmap, 0, 0x81);
	inst=0;
	for (i=0; i<0x80; i++)
		if (instused[i])
		{
			int stat;
			stat=loadpatch(&m->instruments[inst], i, sampused[i], &smps[inst], &m->sampnum);
			fprintf(stderr, "Loaded patch %s\n", m->instruments[inst].name);
			if (stat)
			{
				midClose();
				free(smps);
				return stat;
			}
			m->instruments[inst].prognum=i;
			m->instmap[i]=inst;
			inst++;
		}

	if (instused[0x80] && !midUseFFF) /* fff-todo, blah... */
	{
		struct minstrument *ins;
		uint16_t drums=0;
		uint8_t sn;
		for (i=0; i<0x80; i++)
			if (sampused[0x80][i>>3]&(1<<(i&7)))
				if (midInstrumentNames[i+0x80][0])
					drums++;
		m->instmap[0x80]=inst;
		ins=&m->instruments[inst];
		ins->prognum=0x80;
		ins->sampnum=drums;
		smps[inst]=calloc(sizeof(struct sampleinfo), drums);
		if (!(ins->samples=calloc(sizeof(struct msample), drums)))
		{
			midClose();
			free(smps);
			return errAllocMem;
		}
		memset(ins->note, 0xFF, 0x80);
		strcpy(ins->name, "drums");
		sn=0;
		for (i=0; i<0x80; i++)
			if (sampused[0x80][i>>3]&(1<<(i&7)))
				if (midInstrumentNames[i+0x80][0])
				{
					int stat;
					ins->note[i]=sn;
					stat=addpatch(ins, i+0x80, sn, i, &smps[inst][sn], &m->sampnum);
					fprintf(stderr, "Added patch %s\n", ins->name);
					if (stat)
					{
						midClose();
						return stat;
					}
					sn++;
				}
		inst++;
	}
	free(sampused);

	m->samples=calloc(sizeof(struct sampleinfo), m->sampnum);

	samplenum=0;
	for (i=0; i<inst; i++)
	{
		for (j=0; j<m->instruments[i].sampnum; j++)
			m->samples[samplenum++]=smps[i][j];
		free(smps[i]);
	}
	free(smps);

	midClose();
	return errOk;
}

void mid_reset(struct midifile *mf)
{
	mf->tracks=0;
	mf->instruments=0;
	mf->samples=0;
}

void mid_free(struct midifile *mf)
{
	int i;
	if (mf->tracks)
	{
		for (i=0; i<mf->tracknum; i++)
			if (mf->tracks[i].trk)
				free(mf->tracks[i].trk);
		free(mf->tracks);
	}
	if (mf->instruments)
	{
		for (i=0; i<mf->instnum; i++)
			if (mf->instruments[i].samples)
				free(mf->instruments[i].samples);
		free(mf->instruments);
	}
	if (mf->samples)
	{
		for (i=0; i<mf->sampnum; i++)
			free(mf->samples[i].ptr);
		free(mf->samples);
	}
	mid_reset(mf);
}

void closeFFF(void)
{
	struct FFF_ENVP_LIST *el=envp_list;
	struct FFF_PTCH_LIST *pl=ptch_list;
	struct FFF_DATA_LIST *dl=data_list;

	while (el)
	{
		struct FFF_ENVP_LIST *t=el->next;
		int i;
		for (i=0; i<el->chunk->num_envelopes; i++)
		{
			free(el->chunk->records[i].attack_points);
			free(el->chunk->records[i].release_points);
		}
		free(el->chunk->records);
		free(el->chunk);
		free(el);
		el=t;
	}

	while (pl)
	{
		struct FFF_PTCH_LIST *t=pl->next;
		int i;
		for (i=0; i<pl->chunk->nlayers; i++)
			free(pl->chunk->iw_layer[i].waves);

		free(pl->chunk->iw_layer);
		free(pl->chunk);

		free(pl);;
		pl=t;
	}

	while (dl)
	{
		struct FFF_DATA_LIST *t=dl->next;
		free(dl->chunk);
		free(dl);
		dl=t;
	}
}

static inline int ulaw2linear(uint8_t u_val)
{
	return ulaw_exp_table[u_val];
}

int addpatchFFF( struct minstrument *ins,
                 uint8_t             program,
                 uint8_t             sn,
                 uint8_t             sampnum,
                 struct sampleinfo  *sip,
                 uint16_t           *samplenum)
{
	return 0;
}

/*
 * instrument name table for .FFF files (melodic bank, after GM standard)
 */

static char *gmins_melo[]={
	"Acoustic Piano 1", "Acoustic Piano 2", "Acoustic Piano 3",
	"Honky-tonk", "E-Piano 1", "E-Piano 2", "Harpsichord",
	"Clavinet", "Celesta", "Glockenspiel", "Music Box",
	"Vibraphone", "Marimbaphone", "Xylophone", "Tubular-bell",
	"Santur", "Organ 1", "Organ 2", "Organ 3", "Church Organ",
	"Reed Organ", "Accordion", "Harmonica", "Bandoneon",
	"Nylon-str. Guit.", "Steel-str. Guit.", "Jazz Guitar",
	"Clean Guitar", "Muted Guitar", "Overdrive Guitar",
	"Distortion Guit.", "Guitar Harmonics", "Acoustic Bass",
	"Fingered Bass", "Picked Bass", "Fretless Bass",
	"Slap Bass 1", "Slap Bass 2", "Synth Bass 1", "Synth Bass 2",
	"Violin", "Viola", "Cello", "Contrabass", "Tremolo String",
	"Pizzicato String", "Harp", "Timpani", "Strings",
	"Slow Strings", "Synth Strings 1", "Synth Strings 2", "Choir",
	"Voice Oohs", "SynVox", "Orchestra Hit", "Trumpet",
	"Trombone", "Tuba", "Muted Trumpet", "French Horn", "Brass 1",
	"Synth Brass 1", "Synth Brass 2", "Soprano Saxophone",
	"Alto Saxophone", "Tenor Saxophone", "Baritone Saxophone",
	"Oboe", "English Horn", "Bassoon", "Clarinet", "Piccolo",
	"Flute", "Recorder", "Pan Flute", "Bottle Blow",
	"Shakuhachi", "Whistle", "Ocarina", "Square Wave", "Saw wave",
	"Synth Calliope", "Chiffer Lead", "Charang", "Solo Vox",
	"5th Saw Wave", "Bass & Lead", "Fantasia", "Warm Pad",
	"Polysynth", "Space Voice", "Bowed Glass", "Metal Pad",
	"Halo Pad", "Sweeo Pad", "Ice Rain", "Soundtrack", "Crystal",
	"Atmosphere", "Brightness", "Goblin", "Echo Drops",
	"Star Theme", "Sitar", "Banjo", "Shamisen", "Koto", "Kalima",
	"Bag Pipe", "Fiddle", "Shannai", "Tinkle Bell", "Agogo",
	"Steel Drums", "Woodblock", "Taiko", "Melodic Tom 1",
	"Synth Drum", "Reverse Cymbal", "Guitar FretNoise",
	"Breath Noise", "Seashore", "Bird", "Telephone 1",
	"Helicopter", "Applause", "Gun Shot"
};

int loadpatchFFF(struct minstrument *ins,
                 uint8_t             program,
                 uint8_t            *sampused,
                 struct sampleinfo **smps,
                 uint16_t           *samplenum)
{
	struct FFF_PTCH_LIST *pl=ptch_list;
	struct FFF_PTCH_CHUNK *patch;
	struct FFF_LAYR_CHUNK *layer;
	struct FFF_ENVP_CHUNK *envelope[2];

	int i, e, n, smp;

	while (pl)
	{
		if (pl->chunk->program==program) break;
		pl=pl->next;
	}
	if (!pl)
	{
		fprintf(stderr, "[FFF]: program %d not found!\n", program);
		return errGen;
	}

	patch=pl->chunk;
	layer=patch->iw_layer;       /* todo: more layers?!? don't know how.. (fd) */
	envelope[0]=(struct FFF_ENVP_CHUNK*)layer->penv.ptr;
	envelope[1]=(struct FFF_ENVP_CHUNK*)layer->venv.ptr;

/*	sprintf(ins->name, ".-.-.-", program); */

	if (program<128)
		strcpy(ins->name, gmins_melo[program]);
	else
		sprintf(ins->name, "#%d", program);

	ins->prognum=program;
	ins->sampnum=layer->nwaves;
	ins->samples=calloc(sizeof(struct msample), layer->nwaves);

	*smps=calloc(sizeof(struct sampleinfo), layer->nwaves);

	fprintf(stderr, "loading program %d\n", program);

	for(i=0; i<2; i++)
	{
		if(!(i?layer->venv.ptr:layer->penv.ptr)) continue;
		fprintf(stderr, "%s envelope(s):\n", i?"volume":"pitch");
		fprintf(stderr, "   retrigger: %d\n", envelope[i]->flags);
		fprintf(stderr, "   mode     : %d\n", envelope[i]->mode);
		fprintf(stderr, "   indtype  : %d\n", envelope[i]->indtype);
		for(e=0; e<envelope[i]->num_envelopes; e++)
		{
			struct FFF_ENVELOPE_RECORD *rec=&envelope[i]->records[e];
			fprintf(stderr, "  env #%d: (hirange: %d)\n", e, rec->hirange);
			fprintf(stderr, "   sustain_offset:        %d\n", rec->sustain_offset);
			fprintf(stderr, "   sustain_rate  :        %d\n", rec->sustain_rate);
			fprintf(stderr, "   release_rate  :        %d\n", rec->release_rate);
			fprintf(stderr, "   attack_envelope: \n    next:   ");
			for(n=0; n<rec->nattack; n++)
				fprintf(stderr, "%04d ", rec->attack_points[n].next);
			fprintf(stderr, "\n    r/t :   ");
			for(n=0; n<rec->nattack; n++)
				fprintf(stderr, "%04d ", rec->attack_points[n].rate);
			fprintf(stderr, "\n   release_envelope:\n    next:   ");
			for(n=0; n<rec->nrelease; n++)
				fprintf(stderr, "%04d ", rec->release_points[n].next);
			fprintf(stderr, "\n    r/t :   ");
			for(n=0; n<rec->nrelease; n++)
				fprintf(stderr, "%04d ", rec->release_points[n].rate);
			fprintf(stderr, "\n");
		}
	}

	for (smp=0; smp<layer->nwaves; smp++)
	{
		struct FFF_WAVE_CHUNK *wave=&layer->waves[smp];
		struct msample *s=&ins->samples[smp];
		int freq;
		int n, q;
		char *nf, *nt;
		FILE *dat;
		char filename[PATH_MAX+NAME_MAX];

		s->handle=(*samplenum)++;
		fprintf(stderr, "(3)Loaded sample %x\n", s->handle);
		s->sampnum=smp;

		s->normnote=layer->freq_center<<8;

		freq=(44100.0*1024.0/(float)wave->sample_ratio*(float)getfreq(s->normnote)/(float)1000);

/* ratio = 44100 * (root_freq * 1000) * 1024 / sample_rate / 1000
 * is this really right? (fd)  don't know, but seems to work (but some samples sound mistuned, hmmm...) (ryg)
 */

		for (n=wave->low_note; n<wave->high_note; n++)
			ins->note[n]=smp;

		for (q=0; q<6; q++)
		{
			s->volpos[q]=0;
			s->volrte[q]=0;
		}

		s->volpos[0]=62976;
		s->volrte[0]=250<<8;
		s->volpos[1]=62976;
		s->volrte[1]=0;
		s->end=2;
		s->sustain=1;
		s->sclfac=256;       /* ?!?? */
		s->sclbas=60;

		/* todo: double check them again.
		 * did it (and corrected 'em) (ryg)
		 */
/*		s->tremswp=(layer->tremolo.sweep-10)*2*64/45;
		s->tremrte=((float)layer->tremolo.freq*7.0/3.0+15.0)*65536.0/(300.0*64.0);
		s->tremdep=layer->tremolo.depth*19*4*256/255/2;

		s->vibswp=(layer->vibrato.sweep-10)*2*64/45;
		s->vibrte=((float)layer->vibrato.freq*7.0/3.0+15.0)*65536.0/(300.0*64.0);
		s->vibdep=layer->vibrato.depth*19*4*256/255/2;

    ryg 990125: seems to be incorrect (vibratos glide over the whole fft
                analyser :)  - trying to fix this by combining beisert's
                formulas for PAT with some from GIPC ("a glorious hack")
*/

		s->tremswp=(layer->tremolo.sweep-10)*2*64/45;
		s->tremrte=(((float)layer->tremolo.freq*7.0/3.0)+15.0)*65536.0/(300.0*64);
		s->tremdep=(layer->tremolo.depth*51*256)/(255*7);

		s->vibswp=(layer->vibrato.sweep-10)*2*64/45;
		s->vibrte=(((float)layer->vibrato.freq*7.0/3.0)+15.0)*65536.0/(300.0*64);
		s->vibdep=(layer->vibrato.depth*12*256)/(255*7);

		fprintf(stderr, "   -> %d %d %d, %d %d %d\n", s->tremswp, s->tremrte, s->tremdep,
		                                              s->vibswp, s->vibrte, s->vibdep);

		(*smps)[smp].type =(wave->format&1)?0:mcpSamp16Bit;
		(*smps)[smp].type|=(wave->format&2)?0:mcpSampUnsigned;
		(*smps)[smp].type|=(wave->format&8)?mcpSampLoop:0;
		(*smps)[smp].type|=(wave->format&16)?mcpSampBiDi:0;
/*		(*smps)[smp].type|=(wave->format&32)?mcpSampULaw:0; */
		(*smps)[smp].type|=(wave->format&32)?mcpSamp16Bit:0;

		(*smps)[smp].ptr=calloc(wave->size*((wave->format&1)?1:2), ((wave->format&32)?2:1));
		(*smps)[smp].length=wave->size;
		(*smps)[smp].samprate=freq;
		(*smps)[smp].loopstart=wave->loopstart>>4;
		(*smps)[smp].loopend=wave->loopend>>4;

		(*smps)[smp].sloopstart=0;
		(*smps)[smp].sloopend=0;

		nf=plNoteStr[wave->low_note];
		nt=plNoteStr[wave->high_note];
		sprintf(s->name, "%c%c%c to %c%c%c", nf[0], nf[1], nf[2],
		                                     nt[0], nt[1], nt[2]);

		snprintf(filename, sizeof(filename), "%s%s", cfGetProfileString("midi", "dir", "./"), ((struct FFF_DATA_CHUNK *)wave->data_id.ptr)->filename);

		if (!(dat=fopen(filename, "r")))
		{
			fprintf(stderr, "[FFF]: %s not found.\n", filename);
			free ((*smps)[smp].ptr);
			return errGen;
		}

		fseek(dat, wave->start, SEEK_SET);

		if(wave->format&32) /* ulaw */
		{
			int8_t *temp=calloc(wave->size, 1);
			uint16_t *d=(uint16_t *)(*smps)[smp].ptr;
			int i;
			fread(temp, wave->size, 1, dat);
			for (i=0; i<wave->size; i++)
				((short*)d)[i]=ulaw2linear(temp[i]);
			free(temp);
		} else
			fread((*smps)[smp].ptr, wave->size*((wave->format&1)?1:2), 1, dat);

		fclose(dat);
	}

	return errOk;
}
