/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * WAVPlay assembler routines for amplifying/clipping
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -ss04????   Stian Skjelstad <stian@nixia.no>
 *    -translated assembler to gcc
 *  -ss040908   Stian Skjelstad <stian@nixia.no>
 *    -made assembler optimize-safe for gcc
 */

#include "../config.h"
#include "wavpasm.h"

#include <stdio.h>

#ifdef I386_ASM

void mixClipAlt(unsigned short *dst, const unsigned short *src, unsigned long len, const unsigned short *tab)
{
	int d0, d1, d2;
	__asm__ __volatile__
	(
#ifdef __PIC__
	 	"pushl %%ebx\n"
		"movl  %%edx, %%ebx\n"
#endif
		"  pushl %%ebp\n"
		"  xorl %%edx, %%edx\n"

		"  pushl %%ecx\n" /* macro, exp=4 , blocksize = (1 shl exp) = 16, blockexp=exp*/
		"  andl $15, %%ecx\n" /* blocksize-1 */
		"  subl $16, %%ecx\n" /* blocksize */
		"  negl %%ecx\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%edi\n"
		"  subl %%ecx, %%edi\n"

		"  movl $blockend, %%eax\n"
		"  subl $block, %%eax\n"
		"  shrl $4, %%eax\n"

		"  mulb %%cl\n"
		"  addl $block, %%eax\n"
		"  popl %%ecx\n"
		"  shrl $4, %%ecx\n"
		"  incl %%ecx\n"
		"  jmp *%%eax\n"

		"block:\n"

/*		"  nop\n"
 *		"  nop\n" */
		".macro dummy from=0, to=30\n"
		"  movb 1+\\from(%%esi), %%dl\n"
		"  movl (%%ebx,%%edx,4), %%ebp\n"
		"  movl 1024(%%ebx,%%edx,2), %%eax\n"
		"  movb \\from(%%esi), %%dl\n"
		"  addl (%%ebp,%%edx,2), %%eax\n"
		"  movw %%ax, \\from(%%edi)\n"
		".if \\to-\\from\n"
		" dummy \"(\\from+2)\",\\to\n"
		".endif\n"
		".endm\n"
		" dummy\n"

		"blockend:\n"
		"  addl $32, %%esi\n" /* 128 ??? */
		"  addl $32, %%edi\n" /* 128 ??? */
		"  decl %%ecx\n"
		"  jnz block\n"

		"  popl %%ebp\n"
#ifdef __PIC__
		"movl  %%ebx, %%edx\n"
		"popl %%ebx\n"
		: "=&S"(d0), "=&D"(d1), "=&c"(d2)
		: "0" (src), "1" (dst), "d" (tab), "2" (len)
		: "memory", "eax"
#else
		: "=&S"(d0), "=&D"(d1), "=&c"(d2)
		: "0" (src), "1" (dst), "b" (tab), "2" (len)
		: "memory", "eax", "edx"
#endif

	);
}


void mixClipAlt2(unsigned short *dst, const unsigned short *src, unsigned long len, const unsigned short *tab)
{
	int d0, d1, d2;
	__asm__ __volatile__
	(
#ifdef __PIC__
	 	"pushl %%ebx\n"
		"movl  %%edx, %%ebx\n"
#endif
		"  pushl %%ebp\n"

		"  xorl %%edx, %%edx\n"

		"  pushl %%ecx\n" /* macro, exp=4 , blocksize = (1 shl exp) = 16, blockexp=exp*/
		"  andl $15, %%ecx\n" /* blocksize-1 */
		"  subl $16, %%ecx\n" /* blocksize */
		"  negl %%ecx\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%esi\n"
		"  subl %%ecx, %%edi\n"
		"  subl %%ecx, %%edi\n"
		"  subl %%ecx, %%edi\n"
		"  subl %%ecx, %%edi\n"

		"  movl $blockend2, %%eax\n"
		"  subl $block2, %%eax\n"
		"  shrl $4, %%eax\n"
	
		"  mulb %%cl\n"
		"  addl $block2, %%eax\n"
		"  popl %%ecx\n"
		"  shrl $4, %%ecx\n"
		"  incl %%ecx\n"
		"  jmp *%%eax\n"

		"block2:\n"

/*		"  nop\n"
		"  nop\n" */
		".macro dummy2 from=0, to=60\n"
		"  movb 1+\\from(%%esi), %%dl\n"
		"  movl (%%ebx,%%edx,4), %%ebp\n"
		"  movl 1024(%%ebx,%%edx,2), %%eax\n"
		"  movb \\from(%%esi), %%dl\n"
		"  addl (%%ebp,%%edx,2), %%eax\n"
		"  movw %%ax, \\from(%%edi)\n"
		".if \\to-\\from\n"
		" dummy2 \"(\\from+4)\",\\to\n"
		".endif\n"
		".endm\n"
		" dummy2\n"

		"blockend2:\n"
		"  addl $64, %%esi\n"
		"  addl $64, %%edi\n"
		"  decl %%ecx\n"
		"  jnz block2\n"

		"  popl %%ebp\n"
#ifdef __PIC__
		"movl  %%ebx, %%edx\n"
		"popl %%ebx\n"
		: "=&S"(d0), "=&D"(d1), "=&c"(d2)
		: "0" (src), "1" (dst), "d" (tab), "2" (len)
		: "memory", "eax"
#else
		: "=&S"(d0), "=&D"(d1), "=&c"(d2)
		: "0" (src), "1" (dst), "b" (tab), "2" (len)
		: "memory", "eax", "edx"
#endif

	);
}

#endif
