/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Routines for screen output
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -kb980717   Tammo Hinrichs <kb@nwn.de>
 *    -did a LOT to avoid redundant mode switching (really needed with
 *     today's monitors)
 *    -added color look-up table for redefining the palette
 *    -added MDA output mode for all Hercules enthusiasts out there
 *     (really rocks in Windows 95 background :)
 *  -fd981220   Felix Domke    <tmbinc@gmx.net>
 *    -faked in a LFB-mode (required some other changes)
 *  -doj20020418 Dirk Jagdmann <doj@cubic.org>
 *    -added screenshot routines
 *  -ss040613   Stian Skjelstad <stian@nixi.no>
 *    -more or less killed the entire file due to the linux implementation, but
 *     the generic functions is still kept in this file.
 */

#include "config.h"
#include <string.h>
#include "types.h"
#include "poutput.h"
#include "imsrtns.h"

unsigned char plpalette[256];
int plScrLineBytes;

char *convnum(unsigned long num, char *buf, unsigned char radix, unsigned short len, char clip0)
{
  unsigned short i;
  for (i=0; i<len; i++)
  {
    buf[len-1-i]="0123456789ABCDEF"[num%radix];
    num/=radix;
  }
  buf[len]=0;
  if (clip0)
    for (i=0; i<(len-1); i++)
    {
      if (buf[i]!='0')
        break;
      buf[i]=' ';
    }
  return buf;
}

void writenum(uint16_t *buf, unsigned short ofs, unsigned char attr, unsigned long num, unsigned char radix, unsigned short len, char clip0)
{
  char convbuf[20];
  char *p=(char *)buf+ofs*2;
  char *cp=convbuf+len;
  unsigned short i;
  for (i=0; i<len; i++)
  {
    *--cp="0123456789ABCDEF"[num%radix];
    num/=radix;
  }
  for (i=0; i<len; i++)
  {
    if (clip0&&(convbuf[i]=='0')&&(i!=(len-1)))
    {
      *p++=' ';
      cp++;
    }
    else
    {
      *p++=*cp++;
      clip0=0;
    }
    *p++=attr;
  }
}

void writestring(uint16_t *buf, unsigned short ofs, unsigned char attr, const char *str, unsigned short len)
{
	char *p=((char *)buf)+ofs*2;
	unsigned short i;
	for (i=0; i<len; i++)
	{
		*p++=*str;
		*p++=attr;
		if (*str)
			str++;
	}
}

void writestringattr(uint16_t *buf, unsigned short ofs, const void *str, unsigned short len)
{
	memcpyb((char *)buf+ofs*2, (void *)str, len*2);
}


void markstring(uint16_t *buf, unsigned short ofs, unsigned short len)
{
  short i;
  buf+=ofs;
  for (i=0; i<len; i++)
    *buf++^=0x8000;
}

void fillstr(uint16_t *buf, const unsigned short ofs, const unsigned char chr, const unsigned char attr, unsigned short len)
{
	buf+=ofs;
	while(len)
	{
		*buf=(chr<<8)+attr;
		buf++;
		len--;
	}

}

